import requests
import blankly
from blankly.data import PriceReader


def price_event(price, symbol, state: blankly.StrategyState):
    """ This function will give an updated price every 15 seconds from our definition below """
    state.variables['history'].append(price)
    rsi = blankly.indicators.rsi(state.variables['history'])
    curr_value = state.interface.account[state.base_asset].available
    if rsi[-1] < 30 and not curr_value:
        # Dollar cost average buy
        buy = blankly.trunc(state.interface.cash / price, 5)
        state.interface.market_order(symbol, side='buy', size=buy)
    elif rsi[-1] > 70 and curr_value:
        # Dollar cost average sell
        state.interface.market_order(symbol, side='sell', size=curr_value)


def init(symbol, state: blankly.StrategyState):
    # Download price data to give context to the algo
    state.variables['history'] = state.interface.history(symbol, to=150, return_as='deque',
                                                         resolution=state.resolution)['close']


if __name__ == "__main__":
    # This downloads an example CSV
    data = requests.get(
        'https://firebasestorage.googleapis.com/v0/b/blankly-6ada5.appspot.com/o/demo_data.csv?alt=media&token=acfa5c39-8f08-45dc-8be3-2033dc2b7b28').text
    with open('./price_examples.csv', 'w') as file:
        file.write(data)

    # Run on the keyless exchange, starting at 100k
    exchange = blankly.KeylessExchange(price_reader=PriceReader('./price_examples.csv', 'BTC-USD'))

    # Use our strategy helper
    strategy = blankly.Strategy(exchange)

    # Make the price event function above run every day
    strategy.add_price_event(price_event, symbol='BTC-USD', resolution='1d', init=init)

    # Backtest the strategy
    results = strategy.backtest(start_date=1598377600, end_date=1650067200, initial_values={'USD': 10000})
    print(results)



"""
    Updated CLI for interacting with and uploading blankly models.
    Copyright (C) 2022 Matias Kotlik

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import argparse
import json
import os.path
import shutil
import subprocess
import sys
import tempfile
import traceback
import webbrowser
from pathlib import Path
from typing import Optional
import pkgutil

from questionary import Choice

from blankly.deployment.api import API
from blankly.deployment.deploy import zip_dir, get_python_version
from blankly.deployment.keys import add_key, load_keys, write_keys
from blankly.deployment.login import logout, poll_login, get_token
from blankly.deployment.ui import text, confirm, print_work, print_failure, print_success, select, show_spinner, path
from blankly.deployment.exchange_data import EXCHANGES, Exchange, EXCHANGE_CHOICES, exc_display_name, \
    EXCHANGE_CHOICES_NO_KEYLESS
from blankly.utils.utils import load_deployment_settings, load_user_preferences, load_backtest_preferences
TEMPLATES = {'strategy': {'none': 'none.py',
                          'rsi_bot': 'rsi_bot.py'},
             'screener': {'none': 'none_screener.py',
                          'rsi_screener': 'rsi_screener.py'}}

AUTH_URL = 'https://app.blankly.finance/auth/signin?redirectUrl=/deploy'


def validate_non_empty(text):
    if not text.strip():
        return 'Please enter a value'
    return True


def create_model(api, name, description, model_type, project_id=None):
    with show_spinner('Creating model') as spinner:
        try:
            # Expanded this logic because the other one was broken
            if project_id is not None:
                model = api.create_model(project_id, model_type, name, description)
            else:
                model = api.create_model(api.user_id, model_type, name, description)
        except Exception:
            spinner.fail('Failed to create model')
            raise
        spinner.ok('Created model')

    return model


def ensure_login() -> API:
    # TODO print selected team ?
    api = is_logged_in()
    if api:
        return api
    return launch_login_flow()


def is_logged_in() -> Optional[API]:
    token = get_token()
    if not token:
        return

    # log into deployment api
    try:
        return API(token)
    except Exception:  # TODO
        return


def launch_login_flow() -> API:
    try:
        webbrowser.open_new(AUTH_URL)
        print_work(f'Your browser was opened to {AUTH_URL}. Open the window and login.')
    except Exception:
        print_work(f'Could not find a browser to open. Navigate to {AUTH_URL} and login')

    api = None
    with show_spinner(f'Waiting for login') as spinner:
        try:
            api = poll_login()
        except Exception:
            pass  # we just check for api being valid, poll_login can return None

        if not api:
            spinner.fail('Failed to login')
            sys.exit(1)

        spinner.ok('Logged in')
    return api


def add_key_interactive(exchange: Exchange):
    tld = None
    if exchange.tlds:
        tld = select('What TLD are you using?',
                     [Choice(f'{exchange.name}.{tld}', tld) for tld in exchange.tlds]).unsafe_ask()

    name = text('Give this key a name:', instruction='(Optional)').unsafe_ask()

    saved_data = {}
    for instruction, key in exchange.key_info.items():
        saved_data[key] = text(f'{instruction}:', validate=validate_non_empty).unsafe_ask().strip()
    saved_data['sandbox'] = confirm('Is this testnet/sandbox key?', default=False).unsafe_ask()

    if add_key(exchange, tld, name, saved_data):
        print_success(f'Your API key for {exchange.display_name} was added to this model')
        return tld
    print_failure(f'Your API Key for {exchange.display_name} was not added to the model')
    return False


def init_starter_model(model):
    path = model.get('path', '.')
    url = model.get('url', 'https://github.com/blankly-finance/examples')
    exchange_name = model.get('exchange', 'alpaca')
    model_type = model.get('modelType', 'strategy')

    with tempfile.TemporaryDirectory() as dir:
        with show_spinner('Downloading files') as spinner:
            ret = subprocess.run(['git', 'clone', url, dir],
                                 stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL).returncode
            if ret != 0:
                spinner.fail('Failed to download starter model. Make sure you have `git` installed.')
                return
            for file in os.listdir(os.path.join(dir, path)):
                try:
                    shutil.move(os.path.join(dir, path, file), './')
                except OSError as e:
                    spinner.fail(f'Failed while copying files: {e}. Try again in an empty directory.')
                    return
            spinner.ok('Downloaded files')

    clean_blankly_json()

    exchange = next((e for e in EXCHANGES if e.name == exchange_name), None)

    if exchange and confirm(f'Would you like to add {exchange.display_name} keys to your model?').unsafe_ask():
        add_key_interactive(exchange)

    if confirm('Would you like to connect this model to the Blankly Platform?').unsafe_ask():
        api = ensure_login()
        ensure_model(api)

    print_success('Done!')


def clean_blankly_json():
    with open('blankly.json', 'r') as file:
        data = json.load(file)
    del data['model_id']
    del data['project_id']
    del data['api_key']
    del data['api_pass']
    with open('blankly.json', 'w') as file:
        json.dump(data, file, indent=4)


def blankly_init(args):
    # warn nonempty dir
    files = [f for f in os.listdir() if not f.startswith('.')]
    if files \
            and not confirm('This directory is not empty. Would you like to continue anyways?').unsafe_ask():
        return

    api = None
    if args.model:
        api = ensure_login()
        starters = api.get_starter_models() or []
        model = next((m for m in starters if m['shortName'] == args.model), None)
        if not model:
            print_failure('That starter model doesn\'t exist. Make sure you are typing the name properly.')
            return
        init_starter_model(model)
        return

    exchange = select('What exchange would you like to connect to?', EXCHANGE_CHOICES).unsafe_ask()

    model_type = 'strategy'
    tld = None
    template = None
    if exchange:
        model_type = select('What type of model do you want to create?',
                            [Choice(mt.title(), mt) for mt in TEMPLATES]).unsafe_ask()
        templates = TEMPLATES[model_type]

        template_name = select('What template would you like to use for your new model?', templates).unsafe_ask()
        template = templates[template_name]

        if confirm('Would you like to add keys for this exchange?\n'
                   'You can do this later at any time by running `blankly key add`').unsafe_ask():
            tld = add_key_interactive(exchange)

    api = None
    model = None
    if args.prompt_login and confirm('Would you like to connect this model to the Blankly Platform?').unsafe_ask():
        if not api:
            api = ensure_login()
        model = get_model_interactive(api, model_type)

    with show_spinner('Generating files') as spinner:
        files = [
            ('bot.py', generate_bot_py(exchange, template), False),
            ('backtest.json', generate_backtest_json(exchange), False),
            ('requirements.txt', 'blankly\n', False),
            ('keys.json', generate_keys_json(), True),
            ('blankly.json', generate_blankly_json(api, model, model_type), False),
            ('settings.json', generate_settings_json(tld or 'com'), False)
        ]
        spinner.ok('Generated files')

    for path, data, force_overwrite in files:
        exists = Path(path).exists()
        overwrite_prompt = confirm(f'{path} already exists, would you like to overwrite it?', default=False)
        if force_overwrite or not exists or overwrite_prompt.unsafe_ask():
            with open(path, 'w') as file:
                file.write(data)

    # TODO open on platform WITHOUT STARTING

    print_success('Done! Your model was created. Run `python bot.py` to run a backtest and get started.')


def get_model_interactive(api, model_type):
    create = select('Would you like to create a new model or attach to an existing one?',
                    [Choice('Create new model', True), Choice('Attach to existing model', False)]).unsafe_ask()
    if create:
        default_name = Path.cwd().name  # default name is working dir name
        name = text('Model name?', default=default_name, validate=validate_non_empty).unsafe_ask()
        description = text('Model description?', instruction='(Optional)').unsafe_ask()
        teams = api.list_teams()
        team_id = None
        if teams:
            team_choices = [Choice('Create on my personal account', False)] \
                           + [Choice(team.get('name', team['id']), team['id']) for team in teams]
            team_id = select('What team would you like to create this model under?', team_choices).unsafe_ask()
        return create_model(api, name, description, model_type, team_id)

    with show_spinner('Loading models...') as spinner:
        models = api.list_all_models()
        spinner.ok('Loaded')

    if not models:
        print_failure('You have no models on the platform. Are you logged into the right account?')
        sys.exit()

    model = select('Select an existing model to attach to:',
                   [Choice(get_model_repr(model), model) for model in models]).unsafe_ask()
    if model.get('type', None) != model_type:
        pass  # TODO
    return model


def get_model_repr(model: dict) -> str:
    name = model.get('name', model['id'])
    team = model.get('team', {}).get('name', None)
    if team:
        name = team + ' - ' + name
    return name


def generate_settings_json(tld: str):
    data = load_user_preferences(override_allow_nonexistent=True)

    data['settings']['binance']['binance_tld'] = tld
    data['settings']['ftx']['ftx_tld'] = tld
    data['settings']['ftx_futures']['ftx_tld'] = tld
    return json.dumps(data, indent=4)


def generate_keys_json():
    try:
        with open('keys.json') as file:
            keys = json.load(file)
    except FileNotFoundError:
        keys = {}

    for exchange in EXCHANGES:
        keys[exchange.name] = keys.get(exchange.name, {
            'example-portfolio': {v: '*' * 20 for v in exchange.key_info.values()}
        })
        for portfolio in keys[exchange.name].values():
            portfolio['sandbox'] = portfolio.get('sandbox', False)
    return json.dumps(keys, indent=4)

  
def generate_blankly_json(api: Optional[API], model: Optional[dict], model_type: str, main_script: str = 'bot.py'):
    data = load_deployment_settings()
    data['main_script'] = main_script
    data['type'] = model_type
    data['python_version'] = get_python_version()
    
    if model:
        data['model_id'] = model['id']
        data['project_id'] = model['projectId']

    if api:
        project_id = model['projectId'] if model else api.user_id
        keys = api.generate_keys(project_id, f'Local keys for {data["model_id"]}')
        data['api_key'] = keys['apiKey']
        data['api_pass'] = keys['apiPass']
    return json.dumps(data, indent=4)


def generate_backtest_json(exchange: Optional[Exchange]) -> str:
    currency = exchange.currency if exchange else 'USD'
    data = load_backtest_preferences(override_allow_nonexistent=True)

    data['settings']['quote_account_value_in'] = currency
    return json.dumps(data, indent=4)


def generate_bot_py(exchange: Optional[Exchange], template: str) -> str:
    if not exchange:
        # template should be None
        return pkgutil.get_data('blankly', 'data/templates/keyless.py').decode('utf-8')
    bot_py = pkgutil.get_data('blankly', f'data/templates/{template}').decode('utf-8')
    for pattern, replacement in [
        ('EXCHANGE_NAME', exchange.display_name,),
        ('EXCHANGE_CLASS', exchange.python_class,),
        ('SYMBOL_LIST', "['" + "', '".join(exchange.symbols) + "']",),
        ('SYMBOL', exchange.symbols[0],),
        ('QUOTE_ASSET', exchange.currency)
    ]:
        bot_py = bot_py.replace(pattern, replacement)
    return bot_py


def blankly_login(args):
    if is_logged_in():
        print_success('You are already logged in')
        return

    launch_login_flow()


def blankly_logout(args):
    with show_spinner('Logging out of the Blankly Platform') as spinner:
        try:
            logout()
        except Exception:
            spinner.fail('Failed to logout')
            raise
        spinner.ok('Logged out')


def ensure_model(api: API):
    # create model if it doesn't exist
    try:
        with open('blankly.json', 'r') as file:
            data = json.load(file)
    except FileNotFoundError:
        print_failure('There was no model detected in this directory. Try running `blankly init` to create one')
        sys.exit(1)

    if 'plan' not in data:
        data['plan'] = select('Select a plan:', [Choice(f'{name} - CPU: {info["cpu"]} RAM: {info["ram"]}', name)
                                                 for name, info in api.get_plans('live').items()]).unsafe_ask()

    if 'model_id' not in data or 'project_id' not in data:
        model = get_model_interactive(api, data.get('type', 'strategy'))
        data['model_id'] = model['modelId']
        data['project_id'] = model['projectId']

    if 'api_key' not in data or 'api_pass' not in data:
        keys = api.generate_keys(data['project_id'], f'Local keys for {data["model_id"]}')
        data['api_key'] = keys['apiKey']
        data['api_pass'] = keys['apiPass']

    files = [f for f in os.listdir() if not f.startswith('.')]
    if 'main_script' not in data or data['main_script'].lstrip('./') not in files:
        if 'bot.py' in files:
            data['main_script'] = 'bot.py'
        else:
            data['main_script'] = path(
                'What is the path to your main script/entry point? (Usually bot.py)').unsafe_ask()

    if data['main_script'].lstrip('./') not in files:
        print_failure(
            f'The file {data["main_script"]} could not be found. Please create it or set a different entry point.')
        sys.exit()

    # save model_id and plan back into blankly.json
    with open('blankly.json', 'w') as file:
        json.dump(data, file, indent=4)

    return data


def missing_deployment_files() -> list:
    paths = ['blankly.json', 'keys.json', 'backtest.json', 'requirements.txt', 'settings.json']
    return [path for path in paths if not Path(path).is_file()]


def blankly_deploy(args):
    api = ensure_login()

    data = ensure_model(api)
    for path in missing_deployment_files():
        if not confirm(f'{path} is missing. Are you sure you want to continue?',
                       default=False).unsafe_ask():
            print_failure('Deployment cancelled')
            print_failure(f'You can try `blankly init` to regenerate the {path} file.')
            return

    description = text('Enter a description for this version of the model:').unsafe_ask()

    with show_spinner('Uploading model') as spinner:
        model_path = zip_dir('.', data['ignore_files'])

        params = {
            'file_path': model_path,
            'project_id': data['project_id'],  # set by ensure_model
            'model_id': data['model_id'],  # set by ensure_model
            'version_description': description,
            'python_version': get_python_version(),
            'type_': data.get('type', 'strategy'),
            'plan': data['plan']  # set by ensure_model
        }
        if data['type'] == 'screener':
            params['schedule'] = data['screener']['schedule']

        response = api.deploy(**params)
        if response.get('status', None) == 'success':
            spinner.ok('Model uploaded')
        else:
            spinner.fail('Error: ' + response['error'])


def blankly_add_key(args):
    exchange = select('What exchange would you like to add a key for?', EXCHANGE_CHOICES_NO_KEYLESS).unsafe_ask()
    add_key_interactive(exchange)


def blankly_list_key(args):
    data = load_keys()
    for exchange_name, keys in data.items():
        for name, key_data in keys.items():
            # filter 'empty'
            if any((isinstance(d, str) and '*' in d) for d in key_data.values()):
                continue
            exchange_display_name = exc_display_name(exchange_name)
            print_work(f'{exchange_display_name}: {name}')
            for k, v in key_data.items():
                print_work(f'    {k}: {v}')


def blankly_remove_key(args):
    data = load_keys()

    key_choices = [Choice(f'{exc_display_name(exchange)}: {name}', (exchange, name))
                   for exchange, keys in data.items() for name in keys]
    if not key_choices:
        return
    exchange, key = select('Which key would you like to delete?', key_choices).unsafe_ask()

    del data[exchange][key]
    write_keys(data)


def blankly_key(args):
    func = select('What would you like to do?', [
        Choice('Add a key', blankly_add_key),
        Choice('Show all keys', blankly_list_key),
        Choice('Remove a key', blankly_remove_key)
    ]).unsafe_ask()
    func(args)


def main():
    parser = argparse.ArgumentParser(prog='blankly', description='Blankly CLI & deployment tool')
    subparsers = parser.add_subparsers(required=True)

    init_parser = subparsers.add_parser('init', help='Initialize a new model in the current directory')
    init_parser.add_argument('model', nargs='?', help='select a starter model')
    init_parser.add_argument('-n', '--no-login', action='store_false', dest='prompt_login',
                             help='don\'t prompt to connect to Blankly Platform')
    init_parser.set_defaults(func=blankly_init)

    login_parser = subparsers.add_parser('login', help='Login to the Blankly Platform')
    login_parser.set_defaults(func=blankly_login)

    logout_parser = subparsers.add_parser('logout', help='Logout of the Blankly Platform')
    logout_parser.set_defaults(func=blankly_logout)

    deploy_parser = subparsers.add_parser('deploy', help='Upload this model to the Blankly Platform')
    deploy_parser.set_defaults(func=blankly_deploy)

    key_parser = subparsers.add_parser('key', help='Manage model Exchange API keys')
    key_parser.set_defaults(func=blankly_key)
    key_subparsers = key_parser.add_subparsers()

    key_add_parser = key_subparsers.add_parser('remove', help='Delete API Keys from this model')
    key_add_parser.set_defaults(func=blankly_remove_key)

    key_add_parser = key_subparsers.add_parser('list', help='List API Keys that this model is using')
    key_add_parser.set_defaults(func=blankly_list_key)

    key_add_parser = key_subparsers.add_parser('add', help='Add an API Key to this model')
    key_add_parser.set_defaults(func=blankly_add_key)

    # run the selected command
    args = parser.parse_args()
    try:
        args.func(args)
    except KeyboardInterrupt:
        print_failure('Cancelled by user')
    except Exception:
        print_failure('An error occurred. Traceback:')
        traceback.print_exc()


if __name__ == '__main__':
    main()



"""
    Websocket feeds need to be modular due to the subscription methods, this provides dynamic management for CBP.
    Copyright (C) 2021  Emerson Dove

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import time

import blankly.utils.utils as utils


def switch_type(stream):
    if stream == "ticker":
        return trade, \
               trade_interface, \
               "time,system_time,price,open_24h,volume_24h,low_24h,high_24h,volume_30d,best_bid,best_ask," \
               "last_size\n"
    elif stream == "level2":
        return no_callback, \
               no_callback, \
               ""
    else:
        return no_callback, no_callback, ""


def no_callback(message):
    return message


def trade(received):
    return ', '.join([
        received['time'],
        time.time(),
        received['price'],
        received['open_24h'],
        received['volume_24h'],
        received['low_24h'],
        received['high_24h'],
        received['volume_24h'],
        received['best_bid'],
        received['best_ask'],
        received['last_size'],
    ]) + '\n'


def trade_interface(message):
    """
    Homogenizing with binance's return:

    Args:
        message (dict): Message from the exchange.
    """
    needed = [
        ["symbol", str],
        ["price", float],
        ["time", float],
        ["trade_id", int],
        ["size", float]
    ]

    message['symbol'] = message.pop('product_id')
    message['size'] = message.pop('last_size')
    return utils.isolate_specific(needed, message)



from typing import Union

from blankly.enums import ContractType, PositionMode, TimeInForce, Side, OrderStatus, OrderType
from blankly.exchanges.interfaces.paper_trade.paper_trade import PaperTrade
from blankly.exchanges.interfaces.coinbase_pro.coinbase_pro import CoinbasePro
from blankly.exchanges.interfaces.exchange_interface import ExchangeInterface
from blankly.exchanges.interfaces.okx.okx import Okx
from blankly.exchanges.interfaces.kucoin.kucoin import Kucoin
from blankly.exchanges.interfaces.binance.binance import Binance
from blankly.exchanges.interfaces.alpaca.alpaca import Alpaca
from blankly.exchanges.interfaces.oanda.oanda import Oanda
from blankly.exchanges.interfaces.ftx.ftx import FTX
from blankly.exchanges.abc_base_exchange import ABCBaseExchange
from blankly.exchanges.interfaces.abc_base_exchange_interface import ABCBaseExchangeInterface
from blankly.exchanges.interfaces.binance_futures.binance_futures import BinanceFutures
from blankly.exchanges.interfaces.ftx_futures.ftx_futures import FTXFutures
from blankly.exchanges.futures.futures_exchange import FuturesExchange
from blankly.exchanges.interfaces.futures_exchange_interface import FuturesExchangeInterface
from blankly.exchanges.interfaces.paper_trade.futures.futures_paper_trade_interface import FuturesPaperTradeInterface
from blankly.exchanges.interfaces.paper_trade.paper_trade_interface import PaperTradeInterface
from blankly.exchanges.orders.futures.futures_order import FuturesOrder
from blankly.exchanges.orders.limit_order import LimitOrder
from blankly.exchanges.orders.market_order import MarketOrder
from blankly.utils import utils

import time
import functools
import os

import pytest
from _pytest.python import Metafunc
from _pytest.python_api import approx

from contextlib import contextmanager

# futures exchanges
binance_futures = BinanceFutures(keys_path="./tests/config/keys.json",
                                 preferences_path="./tests/config/settings.json",
                                 portfolio_name="Futures Test Key")
ftx_futures = FTXFutures(keys_path="./tests/config/keys.json",
                         preferences_path="./tests/config/settings.json",
                         portfolio_name="Dotcom Test Account")

# spot exchanges
coinbase_pro = CoinbasePro(portfolio_name="Sandbox Portfolio",
                           keys_path='./tests/config/keys.json',
                           settings_path="./tests/config/settings.json")
okx = Okx(portfolio_name="okx sandbox portfolio",
          keys_path='./tests/config/keys.json',
          settings_path="./tests/config/settings.json")
kucoin = Kucoin(portfolio_name="KC Sandbox Portfolio",
                keys_path='./tests/config/keys.json',
                settings_path="./tests/config/settings.json")
binance = Binance(portfolio_name="Spot Test Key",
                  keys_path='./tests/config/keys.json',
                  settings_path="./tests/config/settings.json")
alpaca = Alpaca(portfolio_name="alpaca test portfolio",
                keys_path='./tests/config/keys.json',
                settings_path="./tests/config/settings.json")
oanda = Oanda(portfolio_name="oanda test portfolio",
              keys_path='./tests/config/keys.json',
              settings_path="./tests/config/settings.json")
ftx = FTX(portfolio_name="Main Account",
          keys_path='./tests/config/keys.json',
          settings_path="./tests/config/settings.json")

FUTURES_EXCHANGES = [
    FuturesPaperTradeInterface(binance_futures.get_type() + '_paper_trade', binance_futures.interface, {'USDT': 1000}),
    binance_futures,
    ftx_futures,
]
SPOT_EXCHANGES = [
    # PaperTradeInterface deferring to the subinterface get_exchange_type causes problems...
    # TODO ?
    PaperTrade(coinbase_pro, {'USD': 1000}),

    # CoinbasePro testnet *sucks*. Trading on it is always fkn broken and they list like 3 symbols.
    # It breaks all the tests.
    # coinbase_pro,

    okx,
    kucoin,
    binance,
    alpaca,
    oanda,
    ftx
]

for exchange in FUTURES_EXCHANGES + SPOT_EXCHANGES:
    # override auto trunc for new tests
    # old tests use the default if auto_truncate is not set, which is False
    exchange.interface.user_preferences['settings']['auto_truncate'] = True


def get_symbols(exchange: ABCBaseExchangeInterface):
    exchange_type = exchange.get_exchange_type()

    # crypto exchanges that use USD
    if exchange_type.startswith('coinbase') \
            or exchange_type.startswith('ftx'):
        return ['BTC-USD']

    # forex
    if exchange_type.startswith('oanda'):
        return ['EUR-USD', 'GBP/CAD']

    # stonks
    if exchange_type.startswith('alpaca'):
        return ['AAPL', 'MSFT']

    # everything else uses USDT
    return ['ETH-USDT', 'BCH-USDT']


def homogeneity_testing(func=None, check_values: bool = False):
    # allow using without arguments
    # this trick brought to you by Python Cookbook
    if func is None:
        return functools.partial(homogeneity_testing, values=check_values)

    results = {}

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        test_name = os.environ.get('PYTEST_CURRENT_TEST')
        test_name = test_name[test_name.find('[') + 1:test_name.find(']')]
        result = func(*args, **kwargs)
        compare_results(results, result, check_values)
        results[test_name] = result
        return result

    return wrapper


def compare_results(previous_outputs: dict, result, check_values: bool):
    for prev_name, prev_results in previous_outputs.items():
        compare_values(prev_name, prev_results, result, check_values)


def compare_values(other_name, other, this, check_values: bool):
    assert type(this) == type(
        other
    ), f'comparing to {other_name}: types do not match. this={this} other={other}'
    # gapped by python 3.7
    # ), f'comparing to {other_name}: types do not match. {this=} {other=}'
    if isinstance(this, list):
        if this and other:
            for val in this[1:]:  # check for self-homogenity
                compare_values('self', val, this[0], check_values)
            for val in other:
                compare_values(other_name, val, this[0], check_values)
    if isinstance(other, dict):
        other.pop('exchange_specific', None)
        this.pop('exchange_specific', None)
        assert this.keys() == other.keys(
        ), f'comparing to {other_name}: dict keys do not match. ' \
           f'symmetric difference: {set(this.keys()).symmetric_difference(other.keys())}'
        for key in other:
            compare_values(other_name, other[key], this[key], check_values)
    elif check_values:
        assert this == other, f'comparing to {other_name}: values are not equal.'


def wait_till_filled(interface: ABCBaseExchangeInterface, order: Union[FuturesOrder, MarketOrder]):
    if isinstance(interface, ExchangeInterface):
        # don't bother, the lack of 'real' order statuses makes this too much of a pain
        time.sleep(1)
        return interface.get_order(order.get_symbol(), order.get_id())

    retries = 0
    res = interface.get_order(order.symbol, order.id)
    while res.status != OrderStatus.FILLED:
        if retries > 2:
            pytest.fail(f"order was not filled in time. status: {res.status}")
        time.sleep(1 << retries)
        retries += 1
        res = interface.get_order(res.symbol, order.id)
    return res


def place_order(interface: ABCBaseExchangeInterface, symbol: str, side: Side, funds: int, reduce_only: bool = False):
    if isinstance(interface, FuturesExchangeInterface):
        product = interface.get_products(symbol)
        price = interface.get_price(symbol)
        precision = product['size_precision']
    else:
        product = next(p for p in interface.get_products() if p['symbol'] == symbol)
        price = interface.get_price(symbol)
        precision = utils.increment_to_precision(product['base_increment'])
    order_size = utils.trunc(funds / price, precision)

    if isinstance(interface, FuturesExchangeInterface):
        order = interface.market_order(symbol, side, order_size, reduce_only=reduce_only)
        assert order.symbol == symbol
        assert 0 <= order.id
        assert order.status in (OrderStatus.OPEN, OrderStatus.FILLED)
        assert order.type == OrderType.MARKET
        assert order.contract_type == ContractType.PERPETUAL
        assert order.side == side
        assert order.limit_price == 0
        assert order.position == PositionMode.BOTH
        assert order.time_in_force == TimeInForce.GTC
    else:
        order = interface.market_order(symbol, side, order_size)
        assert order.get_symbol() == symbol
        # TODO these are not properly homogenized for spot markets
        assert order.get_status()['status'].lower() in ('new', 'open', 'filled', 'done')
        assert order.get_type().lower() == OrderType.MARKET
        assert order.get_size() == approx(order_size)
        assert order.get_side().lower() == side

    res = wait_till_filled(interface, order)

    if isinstance(interface, FuturesExchangeInterface):
        assert res.status == OrderStatus.FILLED
    else:
        assert res.get_status()["status"].lower() in ('closed', 'filled')

    if isinstance(interface, FuturesExchangeInterface):
        if reduce_only:
            # 1% tolerance
            assert res.price <= price * order_size * 1.01
            assert res.size <= order_size * 1.01

            # set size equal so we can compare
            order.size = res.size
        else:
            assert res.price == approx(price * order_size, rel=0.01)
            assert res.size == approx(order_size)

        # set price, status equal so we can compare
        order.price = res.price
        order.status = res.status
        assert order == res
    else:
        assert res.get_price() == approx(price * order_size, rel=0.01)
        assert res.get_size() == approx(order_size, rel=0.01)

    return order


def sell(interface: ABCBaseExchangeInterface, symbol: str, funds: int = 20, reduce_only: bool = False):
    return place_order(interface, symbol, Side.SELL, funds, reduce_only)


def buy(interface: ABCBaseExchangeInterface, symbol: str, funds: int = 20, reduce_only: bool = False):
    return place_order(interface, symbol, Side.BUY, funds, reduce_only)


def close_all(futures_interface: FuturesExchangeInterface):
    for symbol in futures_interface.get_position():
        close_position(futures_interface, symbol)

    for order in futures_interface.get_open_orders():
        futures_interface.cancel_order(order.symbol, order.id)
    assert len(futures_interface.get_position()) == 0


def close_position(interface: ABCBaseExchange, symbol: str):
    """
    Exit position
    Args:
        interface: the interface to sell on
        symbol: the symbol to sell
    """
    if isinstance(interface, FuturesExchangeInterface):
        position = interface.get_position(symbol)
        if not position:
            return
        if position['size'] < 0:
            order = interface.market_order(symbol,
                                           Side.BUY,
                                           position['size'] * -2,
                                           reduce_only=True)
        elif position['size'] > 0:
            order = interface.market_order(symbol,
                                           Side.SELL,
                                           position['size'] * 2,
                                           reduce_only=True)
        else:
            pytest.fail('position size is zero')
        wait_till_filled(interface, order)
        assert interface.get_position(symbol) is None
    else:
        base = utils.get_base_asset(symbol)
        acc = interface.get_account()
        if base not in acc:
            return
        product = next(p for p in interface.get_products() if p['symbol'] == symbol)
        increment = product['base_increment']
        precision = utils.increment_to_precision(increment)
        position = utils.trunc(acc[base]['available'], precision)
        if 0 < position:
            order = interface.market_order(symbol, 'sell', position)
            wait_till_filled(interface, order)
            assert interface.get_account(base)['available'] == approx(0)


@contextmanager
def cancelling_order(interface: ABCBaseExchangeInterface, symbol: str):
    """
    Context manager for placing and cancelling a limit order from within a test.
    Useful for testing paths that require active orders.
    Args:
        interface: The interface to place an order on
        symbol: The symbol to buy

    Returns: The order object as returned from the interface
    """
    product = interface.get_products(symbol)
    price = interface.get_price(symbol)

    # place our limit order
    if isinstance(interface, FuturesExchangeInterface):
        order_size = utils.trunc(100 / price, product['size_precision'])
        limit_price = utils.trunc(price * 0.90, product['price_precision'])
        order = interface.limit_order(symbol, Side.BUY, limit_price, order_size)
        assert order.symbol == symbol
        assert 0 <= order.id
        assert order.size == approx(order_size)
        assert order.status == OrderStatus.OPEN
        assert order.type == OrderType.LIMIT
        assert order.contract_type == ContractType.PERPETUAL
        assert order.side == Side.BUY
        assert order.position == PositionMode.BOTH
        assert order.limit_price == approx(limit_price)
        assert order.time_in_force == TimeInForce.GTC
    else:
        order_size = utils.trunc(100 / price, utils.increment_to_precision(product['base_increment']))
        limit_price = utils.trunc(price * 0.90, 2)  # 2 should be fine? we don't have quote asset increment here.
        order = interface.limit_order(symbol, 'buy', limit_price, order_size)
        assert order.get_symbol() == symbol
        # TODO these are not properly homogenized for spot markets
        assert order.get_status()['status'].lower() in ('new', 'open', 'filled')
        assert order.get_type().lower() == OrderType.MARKET
        assert order.get_size() == approx(order_size)
        assert order.get_side().lower() == Side.BUY
        assert order.get_price() == approx(limit_price)

    yield order

    res = interface.cancel_order(symbol, order.id)

    if isinstance(interface, FuturesExchangeInterface):
        assert res.status == OrderStatus.CANCELED

        assert order != res
        res.status = OrderStatus.OPEN
        assert order == res



"""
    Class to pass in the current Strategy State
    Copyright (C) 2021  Emerson Dove, Brandon Fan

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from blankly.exchanges.interfaces.abc_exchange_interface import ABCExchangeInterface as Interface
from blankly.utils.utils import AttributeDict, get_base_asset, get_quote_asset, format_with_new_line, pretty_print_json


class StrategyState:
    interface: Interface
    variables: AttributeDict
    resolution: int

    """Strategy State"""
    def __init__(self, strategy, variables: AttributeDict, symbol, resolution: [int, float] = None):
        self.strategy = strategy
        self.variables = variables
        self.resolution = resolution
        self.symbol = symbol

        # Base & quotes are conditionally defined
        self.base_asset = None
        self.quote_asset = None
        if isinstance(symbol, str):
            self.base_asset = get_base_asset(symbol)
            self.quote_asset = get_quote_asset(symbol)

    @staticmethod
    def append_bar(history_reference, new_bar: dict):
        history_reference['open'].append(new_bar['open'])
        history_reference['high'].append(new_bar['high'])
        history_reference['low'].append(new_bar['low'])
        history_reference['close'].append(new_bar['close'])
        history_reference['volume'].append(new_bar['volume'])
        history_reference['time'].append(new_bar['time'])

    @property
    def interface(self) -> Interface:
        """
        Get the interface object that the strategy is running on. Use this to interact with your exchange.
        """
        return self.strategy.interface

    @property
    def time(self) -> float:
        """
        Get the time from the strategy.
        This will automatically switch to match the correct times during backtesting
        """
        return self.strategy.time()

    def __str__(self):
        output = ""
        output = format_with_new_line(output, "Symbol: ", self.symbol)

        output = format_with_new_line(output, "Time: ", self.time)

        output = format_with_new_line(output, "Resolution: ", self.resolution)

        output = format_with_new_line(output, "Variables: ")

        output = format_with_new_line(output, pretty_print_json(self.variables, actually_print=False))

        return output



"""
    Moving average function utils
    Copyright (C) 2021 Brandon Fan

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from typing import Any

import pandas as pd
import tulipy as ti

from blankly.indicators.utils import check_series, convert_to_numpy


def ema(data: Any, period: int = 50, use_series=False) -> Any:
    if check_series(data):
        use_series = True
    data = convert_to_numpy(data)
    ema = ti.ema(data, period=period)
    return pd.Series(ema) if use_series else ema


def vwma(data: Any, volume_data: Any, period: int = 50, use_series=False) -> Any:
    if check_series(data):
        use_series = True

    data = convert_to_numpy(data)
    volume_data = convert_to_numpy(volume_data).astype(float)

    vwma = ti.vwma(data, volume_data, period=period)
    return pd.Series(vwma) if use_series else vwma


def wma(data: Any, period: int = 50, use_series=False) -> Any:
    if check_series(data):
        use_series = True
    data = convert_to_numpy(data)
    wma = ti.wma(data, period)
    return pd.Series(wma) if use_series else wma


def zlema(data: Any, period: int = 50, use_series=False) -> Any:
    if check_series(data):
        use_series = True
    data = convert_to_numpy(data)
    zlema = ti.zlema(data, period)
    return pd.Series(zlema) if use_series else zlema


def sma(data: Any, period: int = 50, use_series=False) -> Any:
    """
    Finding the moving average of a dataset
    Args:
        data: (list) A list containing the data you want to find the moving average of
        period: (int) How far each average set should be
    """
    if check_series(data):
        use_series = True
    data = convert_to_numpy(data)
    sma = ti.sma(data, period=period)
    return pd.Series(sma) if use_series else sma


def hma(data: Any, period: int = 50, use_series=False) -> Any:
    if check_series(data):
        use_series = True
    data = convert_to_numpy(data)
    hma = ti.hma(data, period)
    return pd.Series(hma) if use_series else hma


def kaufman_adaptive_ma(data: Any, period: int = 50, use_series=False) -> Any:
    if check_series(data):
        use_series = True
    data = convert_to_numpy(data)
    kama = ti.kama(data, period)
    return pd.Series(kama) if use_series else kama


def trima(data: Any, period: int = 50, use_series=False) -> Any:
    if check_series(data):
        use_series = True
    data = convert_to_numpy(data)
    trima = ti.trima(data, period)
    return pd.Series(trima) if use_series else trima


def macd(data: Any, short_period: int = 12, long_period: int = 26, signal_period: int = 9, use_series=False) -> Any:
    if check_series(data):
        use_series = True
    data = convert_to_numpy(data)
    macd, macd_signal, macd_histogram = ti.macd(data, short_period, long_period, signal_period)
    if use_series:
        df = pd.DataFrame({'macd': macd, 'macd_signal': macd_signal, 'macd_histogram': macd_histogram})
        return df
    return macd, macd_signal, macd_histogram



"""
    Base class for FuturesExchangeInterfaces.
    Provides common functionality for Futures Exchanges.
    Copyright (C) 2022 Matias Kotlik

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import abc
from typing import Union, Optional, List

import blankly.utils.utils as utils
from blankly.enums import MarginType, HedgeMode, PositionMode, OrderType, Side, TimeInForce, ContractType
from blankly.exchanges.interfaces.abc_base_exchange_interface import ABCBaseExchangeInterface
from blankly.exchanges.orders.futures.futures_order import FuturesOrder


class FuturesExchangeInterface(ABCBaseExchangeInterface, abc.ABC):
    exchange_name: str

    def __init__(self,
                 exchange_name,
                 authenticated_api,
                 preferences_path=None):
        self.exchange_name = exchange_name
        self.calls = authenticated_api

        self.user_preferences = utils.load_user_preferences(preferences_path)

        self.exchange_properties = None
        self.available_currencies = {}

        if self.user_preferences['settings']['test_connectivity_on_auth']:
            self.init_exchange()

    @staticmethod
    def to_blankly_symbol(symbol: str):
        return symbol

    @staticmethod
    def to_exchange_symbol(symbol: str):
        return symbol

    def get_exchange_type(self) -> str:
        """Returns the exchange type (ex. 'binance', 'coinbase', 'alpaca')"""
        return self.exchange_name

    @abc.abstractmethod
    def init_exchange(self):
        """Initializes the exchange"""
        pass

    @abc.abstractmethod
    def get_products(self, symbol: str = None) -> dict:
        """Returns a list of all products traded on this exchange"""
        pass

    @abc.abstractmethod
    def get_account(self, symbol: str = None) -> dict:
        """Returns account information, or information for only one `symbol` if one is given."""
        pass

    # TODO this metohd name might need to change to get_position ?
    @abc.abstractmethod
    def get_position(self, symbol: str = None) -> Optional[dict]:
        """Returns position information, or information for only one `symbol` if one is given"""
        pass

    @abc.abstractmethod
    def market_order(self,
                     symbol: str,
                     side: Side,
                     size: float,
                     position: PositionMode = None,
                     reduce_only: bool = None) -> FuturesOrder:
        """Places a market order"""
        pass

    @abc.abstractmethod
    def limit_order(self,
                    symbol: str,
                    side: Side,
                    price: float,
                    size: float,
                    position: PositionMode = None,
                    reduce_only: bool = None,
                    time_in_force: TimeInForce = None) -> FuturesOrder:
        """Places a limit order"""
        pass

    @abc.abstractmethod
    def take_profit_order(self,
                          symbol: str,
                          side: Side,
                          price: float,
                          size: float,
                          position: PositionMode = None) -> FuturesOrder:
        """Place a take-profit order for a position"""
        pass

    @abc.abstractmethod
    def stop_loss_order(self,
                        symbol: str,
                        side: Side,
                        price: float,
                        size: float,
                        position: PositionMode = None) -> FuturesOrder:
        """Place a stop-loss order for a position"""
        pass

    @abc.abstractmethod
    def set_hedge_mode(self, hedge_mode: HedgeMode):
        pass

    @abc.abstractmethod
    def get_hedge_mode(self):
        pass

    @abc.abstractmethod
    def set_leverage(self, leverage: float, symbol: str = None):
        pass

    @abc.abstractmethod
    def get_leverage(self, symbol: str = None) -> float:
        pass

    @abc.abstractmethod
    def set_margin_type(self, symbol: str, type: MarginType):
        pass

    @abc.abstractmethod
    def get_margin_type(self, symbol: str):
        pass

    @abc.abstractmethod
    def cancel_order(self, symbol: str, order_id: int) -> FuturesOrder:
        """Cancels an order"""
        pass

    @abc.abstractmethod
    def get_open_orders(self, symbol: str = None) -> List[FuturesOrder]:
        """Returns all currently open orders, filtered by `symbol` if one is provided."""
        pass

    @abc.abstractmethod
    def get_order(self, symbol: str, order_id: int) -> FuturesOrder:
        """Returns information for the order corresponding to `order_id`"""
        pass

    @abc.abstractmethod
    def get_price(self, symbol: str) -> float:
        """Returns the current price of an asset"""
        pass

    @property
    def account(self) -> dict:
        """Account information"""
        return self.get_account()

    @property
    def positions(self) -> dict:
        """Position information"""
        return self.get_position()

    @property
    def orders(self) -> list:
        """All currently open orders"""
        return self.get_open_orders()

    @property
    def cash(self) -> float:
        """The amount of cash in a portfolio. The currency for 'cash' is set in settings.json"""
        using_setting = self.user_preferences['settings'][
            self.exchange_name]['cash']
        return self.get_account(using_setting)['available']

    @abc.abstractmethod
    def get_funding_rate_history(self, symbol: str, epoch_start: int,
                                 epoch_stop: int) -> list:
        """
        Get the funding rate history between `epoch_start` and `epoch_end`.
        Returns a list of {'rate': int, 'time': int}
        """
        pass

    @abc.abstractmethod
    def get_funding_rate(self, symbol: str) -> float:
        pass

    @abc.abstractmethod
    def get_maker_fee(self) -> float:
        pass

    @abc.abstractmethod
    def get_taker_fee(self) -> float:
        pass

    @abc.abstractmethod
    def get_funding_rate_resolution(self) -> int:
        pass

    @property
    def should_auto_trunc(self):
        return self.user_preferences['settings'].get('auto_truncate', False)



"""
    Inherited class to abstract the similarities between order types
    Copyright (C) 2021  Emerson Dove

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import blankly
from blankly.utils.utils import pretty_print_json


class Order:
    def __init__(self, response, order, interface):
        """
        Limit Order:
        needed = [
            ["symbol", str],    <- similar
            ["id", str],    <- similar
            ["created_at", float],  <- similar
            ["price", float],
            ["size", float],
            ["status", str],    <- similar
            ["time_in_force", str],     <-similar
            ["type", str],  <- similar
            ["side", str]   <- similar
        ]

        Market Order:
        needed = [
            ["symbol", str],
            ["id", str],
            ["created_at", float],
            ["funds", float],
            ["status", str],
            ["type", str],
            ["side", str]
        ]
        """
        self.__response = response
        self.__order = order
        self.interface = interface
        self.exchange = interface.get_exchange_type()

    @staticmethod
    def add_new_line(input_string, *components, newline=True):
        for i in components:
            input_string += str(i)
        if newline:
            input_string += "\n"
        return input_string

    def __str__(self):
        return_string = ""
        return_string = self.add_new_line(return_string, "General Order Parameters: ")

        return_string = self.add_new_line(return_string, "Response: ")
        return_string = self.add_new_line(return_string, pretty_print_json(self.get_response(), actually_print=False))

        return_string = self.add_new_line(return_string, "Full Status: ")
        return_string = self.add_new_line(return_string, pretty_print_json(self.get_status(full=True),
                                                                           actually_print=False))

        return_string = self.add_new_line(return_string, "ID: ", self.get_id())

        return_string = self.add_new_line(return_string, "Symbol: ", self.get_symbol())

        return_string = self.add_new_line(return_string, "Purchase Time: ", self.get_purchase_time())

        return_string = self.add_new_line(return_string, "Type: ", self.get_type())

        return_string = self.add_new_line(return_string, "Side: ", self.get_side())

        return_string = self.add_new_line(return_string, "Size: ", newline=False)
        return_string = self.add_new_line(return_string, self.get_size())

        return return_string

    def get_response(self) -> dict:
        """
        Get the original but parsed response from the exchange.
        """
        return self.__response

    def get_id(self) -> str:
        """
        Get the exchange-set order ID.
        """
        return self.__response["id"]

    def get_symbol(self) -> str:
        """
        Get the asset id of the exchange
        """
        return self.__response["symbol"]

    def get_purchase_time(self) -> float:
        """
        Get when the purchase was created at. This may be set at slightly different points within an
        exchange's matching engine.
        """
        return self.__response["created_at"]

    def get_status(self, full=True) -> dict:
        """
        Calls the exchange with the order id of this purchase and returns the homogenized output
        Args:
            full: Set this to True to receive the all the order details. Default is False, which returns just the status
            field as (example) {"status:" "NEW"}
        """
        if full:
            return self.interface.get_order(self.__order["symbol"], self.get_id())
        else:
            return {"status": self.interface.get_order(self.__order["symbol"], self.get_id())["status"]}

    def get_type(self) -> str:
        """
        Get order type, such as market/limit/stop.
        """
        return self.__response["type"]

    def get_side(self) -> str:
        """
        Get the order side such as buy/sell as a str.
        """
        return self.__response["side"]

    def annotate(self, annotation: str) -> None:
        """
        Add an annotation to the order on the platform
        """
        return blankly.reporter.annotate_order(self.get_id(), annotation)

    def get_size(self) -> float:
        """
        Get the size of order, Ex: .004 bitcoin
        """
        return self.__response["size"]



"""
    A reader for handling a variety of data formats
    Copyright (C) 2022  Emerson Dove

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import json
import pandas as pd
from enum import Enum

from blankly.exchanges.interfaces.futures_exchange_interface import FuturesExchangeInterface


class DataTypes(Enum):
    ohlcv_csv = 0
    tick_csv = 1
    ohlcv_json = 2
    event_json = 3


"""
This class seeks to solve these problems:

- Read in custom price data for keyless exchange
- Read in event-based
"""


class FileTypes(Enum):
    csv = 'csv'
    json = 'json'


class DataReader:
    @staticmethod
    def _check_length(df: pd.DataFrame, identifier: str):
        try:
            assert (len(df) > 2)
        except AssertionError:
            raise AssertionError(f"Must give data with at least 2 rows in {identifier}.")

    @staticmethod
    def __is_price_data(data_type: DataTypes) -> bool:
        return data_type == DataTypes.ohlcv_json or data_type == DataTypes.ohlcv_csv

    @property
    def data(self):
        return self._internal_dataset

    def _write_dataset(self, contents: dict, key: str, required_columns: tuple):
        try:
            # Check if all the keys exist for each part of the dictionary
            assert (all(key in contents for key in required_columns))
        except AssertionError:
            raise AssertionError(f"Must have at least these columns: {required_columns} for {key}")
        self._internal_dataset[key] = pd.DataFrame.from_dict(contents)

    def __init__(self, data_type: [DataTypes, str]):
        self._internal_dataset = {}
        self.type: DataTypes = data_type
        self.price_data: bool = self.__is_price_data(data_type)


class __FormatReader(DataReader):
    def __init__(self, data_type):
        super().__init__(data_type)

    @staticmethod
    def _convert_to_list(file_path, symbol):
        """
        Primarily for CSV prices, this allows single or multi sets of symbols to be converted into matching arrays
        """
        # Turn it into a list
        if isinstance(file_path, str):
            file_paths = [file_path]
        else:
            file_paths = file_path

        if isinstance(symbol, str):
            symbols = [symbol]
        else:
            # Could be None still
            symbols = symbol

        return file_paths, symbols

    def _parse_csv_prices(self, file_paths: list, symbols: list, columns: set) -> None:
        if symbols is None:
            raise LookupError("Must pass one or more symbols to identify the csv files")
        if len(file_paths) != len(symbols):
            raise LookupError(f"Mismatching symbol & file path lengths, got {len(file_paths)} and {len(symbols)} for "
                              f"file paths and symbol lengths.")

        for index in range(len(file_paths)):
            # Load the file via pandas
            contents = pd.read_csv(file_paths[index])

            self._check_length(contents, file_paths[index])

            # Check if its contained
            assert (columns.issubset(contents.columns))

            # Now push it directly into the dataset and sort by time
            self._internal_dataset[symbols[index]] = contents.sort_values('time')

    def _parse_json_prices(self, file_paths: list, keys: tuple) -> None:
        for file in file_paths:
            # Load the contents into json first because it has a ton of symbols
            contents = json.loads(open(file).read())

            for symbol in contents:
                self._write_dataset(contents[symbol], symbol, keys)

                self._check_length(self._internal_dataset[symbol], file)

                # Ensure that the dataframe is sorted by time
                self._internal_dataset[symbol] = self._internal_dataset[symbol].sort_values('time')


class PriceReader(__FormatReader):
    @staticmethod
    def __associate(file_paths: list) -> str:
        file_type = ''

        def complain_if_different(ending_: str, type_: str):
            nonlocal file_type
            if file_type == '':
                file_type = type_
            elif file_type is not None and file_type != ending_:
                raise LookupError("Cannot pass both csv files and json files into a single constructor.")

        for file_path in file_paths:
            if file_path[-3:] == 'csv':
                complain_if_different('csv', FileTypes.csv.value)
            elif file_path[-4:] == 'json':
                # In this instance the symbols should be None
                complain_if_different('json', FileTypes.json.value)
            else:
                raise LookupError(f"Unknown filetype for {file_path}")

        return file_type

    def _guess_resolutions(self):
        for symbol in self._internal_dataset:
            # Get the time diff
            time_series: pd.Series = self._internal_dataset[symbol]['time']
            time_dif = time_series.diff()

            # Now find the most common difference and use that
            if symbol not in self.prices_info:
                self.prices_info[symbol] = {}

            # Store the resolution start time and end time of each dataset
            self.prices_info[symbol]['resolution'] = int(time_dif.value_counts().idxmax())
            self.prices_info[symbol]['start_time'] = time_series.iloc[0]
            self.prices_info[symbol]['stop_time'] = time_series.iloc[-1]

    def __init__(self, file_path: [str, list], symbol: [str, list]):
        """
        Read in a new custom price dataset in either json or csv format

        Args:
            file_path (str or list): A single file path or list of filepaths pointing to a set of price data
            symbol (str or list): Only required if using .csv files. These must match in index to the symbol that the
             csv file path corresponds to. The CSV files also must have at least 2 rows of data in them.

            symbol (str or list): Pass the symbol or symbols that the file paths correspond to. One file path and one
             symbol can be passed as a non list but multiple can be passed as lists in both arguments. Just make sure
             that the symbol indices match on both arguments
        """
        file_paths, symbols = self._convert_to_list(file_path, symbol)

        # Empty dict to store the resolutions of the inputs by symbol
        self.prices_info = {}

        try:
            assert (len(symbols) == len(set(symbols)))
        except AssertionError:
            raise AssertionError("Cannot use duplicate symbols for one price reader. Please use multiple price readers"
                                 " to read in different datasets of the same symbol.")

        data_type = self.__associate(file_paths)

        super().__init__(data_type)

        if data_type == FileTypes.json.value:
            self._parse_json_prices(file_paths, ('open', 'high', 'low', 'close', 'volume', 'time'))
        elif data_type == FileTypes.csv.value:
            self._parse_csv_prices(file_paths, symbols, {'open', 'high', 'low', 'close', 'volume', 'time'})
        else:
            raise LookupError("No parsing written for input type.")

        self._guess_resolutions()


class EventReader(DataReader):
    def __init__(self, event_type: str, events: dict):
        super().__init__(DataTypes.event_json)
        if events:
            time, data = zip(*events.items())
            self._write_dataset({'time': time, 'data': data}, event_type, ('time', 'data'))


class JsonEventReader(DataReader):
    def __parse_json_events(self, file_path):
        contents = json.loads(open(file_path).read())

        for event_type in contents:
            self._write_dataset(contents[event_type], event_type, ('time', 'data'))

    def __init__(self, file_path):
        super().__init__(DataTypes.event_json)
        try:
            assert file_path[-4:] == 'json'
        except AssertionError:
            raise AssertionError(f"The filepath did not have a \'json\' ending - got: {file_path[-4:]}")

        self.__parse_json_events(file_path)


class FundingRateEventReader(EventReader):
    def __init__(self, symbol: str, start: int, stop: int, interface: FuturesExchangeInterface):
        history = interface.get_funding_rate_history(symbol, start, stop)
        history = {ev['time']: {'symbol': symbol, 'rate': ev['rate']} for ev in history}
        super().__init__('__blankly__funding_rate', history)


class TickReader(__FormatReader):
    def __init__(self, file_path: [str, list], symbol: [str, list] = None):
        super().__init__(DataTypes.tick_csv)
        file_paths, symbols = self._convert_to_list(file_path, symbol)

        try:
            assert file_path[-3:] == 'csv'
        except AssertionError:
            raise AssertionError(f"The filepath did not have a \'csv\' ending - got: {file_path[-3:]}")
        self._parse_csv_prices(file_paths, symbols, {'time', 'price'})



"""
    Alpaca websocket homogenization definitions.
    Copyright (C) 2021  Emerson Dove

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
from alpaca_trade_api.entity_v2 import trade_mapping_v2
from msgpack.ext import Timestamp

from blankly.utils.utils import isolate_specific, rename_to


def parse_alpaca_timestamp(value: Timestamp):
    return value.seconds + (value.nanoseconds * float(1e-9))


def alpaca_remapping(dictionary: dict, mapping: dict):
    # From https://stackoverflow.com/a/67573821/8087739
    try:
        dictionary.pop('T')
    except KeyError:
        pass
    return {mapping[k]: dictionary[k] for k in dictionary}


def switch_type(stream):
    if stream == "trades":
        return trades_logging, \
               trades_interface, \
               "id,symbol,conditions,exchange,price,size,timestamp,tape\n"
    elif stream == "quotes":
        return quotes_logging, \
               no_callback, \
               "symbol,ask_exchange,ask_price,ask_size,bid_exchange,bid_price,bid_size,conditions,timestamp,tape\n"
    else:
        return no_logging_callback, no_callback, ""


def no_callback(message):
    return message


def no_logging_callback(message):
    response = ""
    for i in list(message.keys()):
        response += str(message[i])
        response += ','

    response += '\n'
    return response


def trades_logging(message: dict):
    return str(str(message["i"]) + "," +
               str(message["S"]) + "," +
               str(message["c"]) + "," +
               str(message["x"]) + "," +
               str(message["p"]) + "," +
               str(message["s"]) + "," +
               str(message["t"]) + "," +
               str(message["z"]) + "\n")


def trades_interface(message):
    """
    Similar ticks with coinbase pro
    {
        'type': 'ticker',
        'symbol': 'BTC-USD',
        'price': '50141.55',
        'time': 1619286924.397969,
        'trade_id': 160775277,
        'size': .35324
    }
    """
    message = alpaca_remapping(message, trade_mapping_v2)
    message['time'] = message.pop('timestamp')

    renames = [
        ["id", "trade_id"],
    ]

    message = rename_to(renames, message)

    needed = [
        ["symbol", str],
        ["price", float],
        ["time", float],
        ["trade_id", int],
        ["size", float]
    ]
    return isolate_specific(needed, message)


def quotes_logging(message):
    return str(str(message["S"]) + "," +
               str(message["ax"]) + "," +
               str(message["ap"]) + "," +
               str(message["as"]) + "," +
               str(message["bx"]) + "," +
               str(message["bp"]) + "," +
               str(message["bs"]) + "," +
               str(message["c"]) + "," +
               str(message["t"]) + "," +
               str(message["z"]) + "\n")



"""
    Implementation for Okx Websockets
    Copyright (C) 2022  Emerson Dove

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import json
import traceback

import blankly.exchanges.interfaces.okx.okx_websocket_utils as websocket_utils
from blankly.exchanges.interfaces.websocket import Websocket
from blankly.utils.utils import info_print


class Tickers(Websocket):
    def __init__(self, symbol, stream, log=None,
                 pre_event_callback=None, initially_stopped=False, websocket_url="wss://ws.okx.com:8443/ws/v5/public",
                 **kwargs):
        """
        Create and initialize the ticker
        Args:
            symbol: Currency to initialize on such as "BTC-USD"
            log: Fill this with a path to a log file that should be created
            websocket_url: Default websocket URL feed.
        """
        self.__logging_callback, self.__interface_callback, log_message = websocket_utils.switch_type(stream)

        super().__init__(symbol, stream, log, log_message, websocket_url, pre_event_callback, kwargs)

        self.__pre_event_callback_filled = False

        # Start the websocket
        if not initially_stopped:
            self.start_websocket(
                self.on_open,
                self.on_message,
                self.on_error,
                self.on_close,
                self.read_websocket
            )
        self.checked = False

    def read_websocket(self):
        # Main thread to sit here and run
        self.ws.run_forever()

    def on_message(self, ws, message):
        received_dict = json.loads(message)
        if len(received_dict) == 2 and self.checked is not True:
            info_print(f"Subscribed to {received_dict['arg']['channel']}")
            self.checked = True
            return
        if (self.pre_event_callback is not None) and (not self.__pre_event_callback_filled) and \
                (self.stream == "books"):
            if received_dict['action'] == 'snapshot':
                try:
                    self.pre_event_callback(received_dict)
                except Exception:
                    traceback.print_exc()

            self.__pre_event_callback_filled = True
            return

        self.most_recent_time = received_dict['data'][0]["ts"]
        received_dict['data'][0]["ts"] = self.most_recent_time
        self.time_feed.append(self.most_recent_time)
        self.log_response(self.__logging_callback, received_dict['data'][0])

        # Manage price events and fire for each manager attached
        if self.stream == "books":
            interface_message = self.__interface_callback(received_dict)
        else:  # self.stream == 'tickers':
            interface_message = self.__interface_callback(received_dict['data'][0])
        self.ticker_feed.append(interface_message)
        self.most_recent_tick = interface_message

        try:
            for i in self.callbacks:
                i(interface_message, **self.kwargs)
        except Exception as e:
            info_print(e)
            traceback.print_exc()

        self.message_count += 1

    def on_error(self, ws, error):
        info_print(error)

    def on_close(self, ws):
        # This repeats the close behavior just in case something happens
        pass

    def on_open(self, ws):
        request = json.dumps({
            'op': 'subscribe',
            'args': [{
                'channel': self.stream,
                'instId': self.symbol,
            }]
        })
        ws.send(request)

    def restart_ticker(self):
        self.start_websocket(
            self.on_open,
            self.on_message,
            self.on_error,
            self.on_close,
            self.read_websocket
        )



from blankly.exchanges.futures.futures_exchange import FuturesExchange
from blankly.exchanges.interfaces.futures_exchange_interface import FuturesExchangeInterface
from blankly.exchanges.interfaces.paper_trade.backtesting_wrapper import BacktestingWrapper
from blankly.exchanges.interfaces.paper_trade.futures.futures_paper_trade_interface import FuturesPaperTradeInterface


class FuturesPaperTrade(FuturesExchange):
    def __init__(self,
                 exchange,
                 portfolio_name: str = None,
                 preferences_path: str = None,
                 initial_account_values: dict = None):
        super().__init__("futures_paper_trade", portfolio_name, preferences_path)

        self.exchange = exchange
        self._interface = FuturesPaperTradeInterface(exchange.get_type(), self.exchange.interface,
                                                     initial_account_values)

    @property
    def calls(self):
        return self.exchange.calls

    @property
    def interface(self) -> FuturesExchangeInterface:
        return self._interface



"""
    Time building utils function for easy second building
    Copyright (C) 2021  Emerson Dove

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from typing import Union
from numpy import int64, int32


def build_second() -> int:
    return 1


def build_minute() -> int:
    return build_second() * 60


def build_hour() -> int:
    return build_minute() * 60


def build_day() -> int:
    return build_hour() * 24


def build_week() -> int:
    return build_day() * 7


def build_month() -> int:
    return build_day() * 30


def build_year() -> int:
    return build_day() * 365


def build_decade() -> int:
    return build_year() * 10


def build_century() -> int:
    return build_decade() * 10


def build_millennium() -> int:
    return build_century() * 10


def time_interval_to_seconds(interval: Union[str, float]) -> float:
    """
    Extract the number of seconds in an interval string
    """
    if isinstance(interval, float) or isinstance(interval, int) or isinstance(interval, int64) or \
            isinstance(interval, int32):
        return float(interval)
    # Extract intervals
    if 'mo' in interval:
        interval = interval.replace('mo', 'M')
    try:
        magnitude = int(interval[:-1])
    except ValueError:
        raise ValueError("Invalid time interval definition.")
    unit = interval[-1]

    # Switch units
    if unit == "s":
        base_unit = build_second()
    elif unit == "m":
        base_unit = build_minute()
    elif unit == "h":
        base_unit = build_hour()
    elif unit == "d":
        base_unit = build_day()
    elif unit == "w":
        base_unit = build_week()
    elif unit == "M":
        base_unit = build_month()
    elif unit == "y":
        base_unit = build_year()
    elif unit == "D":
        base_unit = build_decade()
    elif unit == "c":
        base_unit = build_century()
    elif unit == "l":
        base_unit = build_millennium()
    else:
        raise ValueError("Invalid time interval definition.")

    # Scale by the magnitude
    return float(base_unit * magnitude)


def number_interval_to_string(interval: int) -> str:
    """
    This function converts integer intervals into string intervals

    Example: 3600 -> 1h

    Args:
        interval: An integer representing the conversion time
    """
    times = {
        'mo': build_month(),
        'wk': build_week(),
        'd': build_day(),
        'h': build_hour(),
        'm': build_minute(),
        's': build_second()
    }

    unit = None

    for i in times:
        if interval % times[i] == 0:
            unit = i
            break

    magnitude = int(interval/times[unit])

    return f'{magnitude}{unit}'



import requests
import hmac
import base64
import datetime
# import time
import json

CONTENT_TYPE = 'Content-Type'
OK_ACCESS_KEY = 'OK-ACCESS-KEY'
OK_ACCESS_SIGN = 'OK-ACCESS-SIGN'
OK_ACCESS_TIMESTAMP = 'OK-ACCESS-TIMESTAMP'
OK_ACCESS_PASSPHRASE = 'OK-ACCESS-PASSPHRASE'

ACEEPT = 'Accept'
COOKIE = 'Cookie'
LOCALE = 'Locale='

APPLICATION_JSON = 'application/json'

GET = "GET"
POST = "POST"

SERVER_TIMESTAMP_URL = '/api/v5/public/time'

# account
POSITION_RISK = '/api/v5/account/account-position-risk'
ACCOUNT_INFO = '/api/v5/account/balance'
POSITION_INFO = '/api/v5/account/positions'
BILLS_DETAIL = '/api/v5/account/bills'
BILLS_ARCHIVE = '/api/v5/account/bills-archive'
ACCOUNT_CONFIG = '/api/v5/account/config'
POSITION_MODE = '/api/v5/account/set-position-mode'
SET_LEVERAGE = '/api/v5/account/set-leverage'
MAX_TRADE_SIZE = '/api/v5/account/max-size'
MAX_AVAIL_SIZE = '/api/v5/account/max-avail-size'
ADJUSTMENT_MARGIN = '/api/v5/account/position/margin-balance'
GET_LEVERAGE = '/api/v5/account/leverage-info'
MAX_LOAN = '/api/v5/account/max-loan'
FEE_RATES = '/api/v5/account/trade-fee'
INTEREST_ACCRUED = '/api/v5/account/interest-accrued'
INTEREST_RATE = '/api/v5/account/interest-rate'
SET_GREEKS = '/api/v5/account/set-greeks'
ISOLATED_MODE = '/api/v5/account/set-isolated-mode'
MAX_WITHDRAWAL = '/api/v5/account/max-withdrawal'
ACCOUNT_RISK = '/api/v5/account/risk-state'
BORROW_REPAY = '/api/v5/account/borrow-repay'
BORROW_REPAY_HISTORY = '/api/v5/account/borrow-repay-history'
INTEREST_LIMITS = '/api/v5/account/interest-limits'
SIMULATED_MARGIN = '/api/v5/account/simulated_margin'
GREEKS = '/api/v5/account/greeks'

# funding
DEPOSIT_ADDRESS = '/api/v5/asset/deposit-address'
GET_BALANCES = '/api/v5/asset/balances'
FUNDS_TRANSFER = '/api/v5/asset/transfer'
TRANSFER_STATE = '/api/v5/asset/transfer-state'
WITHDRAWAL_COIN = '/api/v5/asset/withdrawal'
DEPOSIT_HISTORIY = '/api/v5/asset/deposit-history'
WITHDRAWAL_HISTORIY = '/api/v5/asset/withdrawal-history'
CURRENCY_INFO = '/api/v5/asset/currencies'
PURCHASE_REDEMPT = '/api/v5/asset/purchase_redempt'
BILLS_INFO = '/api/v5/asset/bills'
PIGGY_BALANCE = '/api/v5/asset/piggy-balance'
DEPOSIT_LIGHTNING = '/api/v5/asset/deposit-lightning'
WITHDRAWAL_LIGHTNING = '/api/v5/asset/withdrawal-lightning'
ASSET_VALUATION = '/api/v5/asset/asset-valuation'
SET_LENDING_RATE = '/api/v5/asset/set-lending-rate'
LENDING_HISTORY = '/api/v5/asset/lending-history'
LENDING_RATE_HISTORY = '/api/v5/asset/lending-rate-history'
LENDING_RATE_SUMMARY = '/api/v5/asset/lending-rate-summary'

# Market Data
TICKERS_INFO = '/api/v5/market/tickers'
TICKER_INFO = '/api/v5/market/ticker'
INDEX_TICKERS = '/api/v5/market/index-tickers'
ORDER_BOOKS = '/api/v5/market/books'
MARKET_CANDLES = '/api/v5/market/candles'
HISTORY_CANDLES = '/api/v5/market/history-candles'
INDEX_CANSLES = '/api/v5/market/index-candles'
MARKPRICE_CANDLES = '/api/v5/market/mark-price-candles'
MARKET_TRADES = '/api/v5/market/trades'
VOLUMNE = '/api/v5/market/platform-24-volume'
ORACLE = '/api/v5/market/oracle'
Components = '/api/v5/market/index-components'
EXCHANGE_RATE = '/api/v5/market/exchange-rate'

# Public Data
INSTRUMENT_INFO = '/api/v5/public/instruments'
DELIVERY_EXERCISE = '/api/v5/public/delivery-exercise-history'
OPEN_INTEREST = '/api/v5/public/open-interest'
FUNDING_RATE = '/api/v5/public/funding-rate'
FUNDING_RATE_HISTORY = '/api/v5/public/funding-rate-history'
PRICE_LIMIT = '/api/v5/public/price-limit'
OPT_SUMMARY = '/api/v5/public/opt-summary'
ESTIMATED_PRICE = '/api/v5/public/estimated-price'
DICCOUNT_INTETEST_INFO = '/api/v5/public/discount-rate-interest-free-quota'
SYSTEM_TIME = '/api/v5/public/time'
LIQUIDATION_ORDERS = '/api/v5/public/liquidation-orders'
MARK_PRICE = '/api/v5/public/mark-price'
TIER = '/api/v5/public/position-tiers'
INTEREST_LOAN = '/api/v5/public/interest-rate-loan-quota'
UNDERLYING = '/api/v5/public/underlying'
VIP_INTEREST_RATE_LOAN_QUOTA = '/api/v5/public/vip-interest-rate-loan-quota'

# TRADING DATA
SUPPORT_COIN = '/api/v5/rubik/stat/trading-data/support-coin'
TAKER_VOLUME = '/api/v5/rubik/stat/taker-volume'
MARGIN_LENDING_RATIO = '/api/v5/rubik/stat/margin/loan-ratio'
LONG_SHORT_RATIO = '/api/v5/rubik/stat/contracts/long-short-account-ratio'
CONTRACTS_INTEREST_VOLUME = '/api/v5/rubik/stat/contracts/open-interest-volume'
OPTIONS_INTEREST_VOLUME = '/api/v5/rubik/stat/option/open-interest-volume'
PUT_CALL_RATIO = '/api/v5/rubik/stat/option/open-interest-volume-ratio'
OPEN_INTEREST_VOLUME_EXPIRY = '/api/v5/rubik/stat/option/open-interest-volume-expiry'
INTEREST_VOLUME_STRIKE = '/api/v5/rubik/stat/option/open-interest-volume-strike'
TAKER_FLOW = '/api/v5/rubik/stat/option/taker-block-volume'

# TRADE
PLACR_ORDER = '/api/v5/trade/order'
BATCH_ORDERS = '/api/v5/trade/batch-orders'
CANAEL_ORDER = '/api/v5/trade/cancel-order'
CANAEL_BATCH_ORDERS = '/api/v5/trade/cancel-batch-orders'
AMEND_ORDER = '/api/v5/trade/amend-order'
AMEND_BATCH_ORDER = '/api/v5/trade/amend-batch-orders'
CLOSE_POSITION = '/api/v5/trade/close-position'
ORDER_INFO = '/api/v5/trade/order'
ORDERS_PENDING = '/api/v5/trade/orders-pending'
ORDERS_HISTORY = '/api/v5/trade/orders-history'
ORDERS_HISTORY_ARCHIVE = '/api/v5/trade/orders-history-archive'
ORDER_FILLS = '/api/v5/trade/fills'
ORDERS_FILLS_HISTORY = '/api/v5/trade/fills-history'
PLACE_ALGO_ORDER = '/api/v5/trade/order-algo'
CANCEL_ALGOS = '/api/v5/trade/cancel-algos'
Cancel_Advance_Algos = '/api/v5/trade/cancel-advance-algos'
ORDERS_ALGO_OENDING = '/api/v5/trade/orders-algo-pending'
ORDERS_ALGO_HISTORY = '/api/v5/trade/orders-algo-history'

# SubAccount
BALANCE = '/api/v5/account/subaccount/balances'
BILLs = '/api/v5/asset/subaccount/bills'
DELETE = '/api/v5/users/subaccount/delete-apikey'
RESET = '/api/v5/users/subaccount/modify-apikey'
CREATE = '/api/v5/users/subaccount/apikey'
WATCH = '/api/v5/users/subaccount/apikey'
VIEW_LIST = '/api/v5/users/subaccount/list'
SUBACCOUNT_TRANSFER = '/api/v5/asset/subaccount/transfer'
ENTRUST_SUBACCOUNT_LIST = '/api/v5/users/entrust-subaccount-list'

# Broker
BROKER_INFO = '/api/v5/broker/nd/info'
CREATE_SUBACCOUNT = '/api/v5/broker/nd/create-subaccount'
DELETE_SUBACCOUNT = '/api/v5/broker/nd/delete-subaccount'
SUBACCOUNT_INFO = '/api/v5/broker/nd/subaccount-info'
SET_SUBACCOUNT_LEVEL = '/api/v5/broker/nd/set-subaccount-level'
SET_SUBACCOUNT_FEE_REAT = '/api/v5/broker/nd/set-subaccount-fee-rate'
SUBACCOUNT_DEPOSIT_ADDRESS = '/api/v5/asset/broker/nd/subaccount-deposit-address'
SUBACCOUNT_DEPOSIT_HISTORY = '/api/v5/asset/broker/nd/subaccount-deposit-history'
REBATE_DAILY = '/api/v5/broker/nd/rebate-daily'
# BROKER_INFO = '/api/v5/broker/nd/info' Broker 获取充值地址文档无法打开，预留位置

# Convert
GET_CURRENCIES = '/api/v5/asset/convert/currencies'
GET_CURRENCY_PAIR = '/api/v5/asset/convert/currency-pair'
ESTIMATE_QUOTE = '/api/v5/asset/convert/estimate-quote'
CONVERT_TRADE = '/api/v5/asset/convert/trade'
CONVERT_HISTORY = '/api/v5/asset/convert/history'

# status
STATUS = '/api/v5/system/status'


def sign(message, secretKey):
    mac = hmac.new(bytes(secretKey, encoding='utf8'), bytes(message, encoding='utf-8'), digestmod='sha256')
    d = mac.digest()
    return base64.b64encode(d)


def pre_hash(timestamp, method, request_path, body):
    return str(timestamp) + str.upper(method) + request_path + body


def get_header(api_key, sign_, timestamp, passphrase, flag):
    header = dict()
    header[CONTENT_TYPE] = APPLICATION_JSON
    header[OK_ACCESS_KEY] = api_key
    header[OK_ACCESS_SIGN] = sign_
    header[OK_ACCESS_TIMESTAMP] = str(timestamp)
    header[OK_ACCESS_PASSPHRASE] = passphrase
    header['x-simulated-trading'] = flag
    return header


def parse_params_to_str(params):
    url = '?'
    for key, value in params.items():
        url = url + str(key) + '=' + str(value) + '&'
    return url[0:-1]


def get_timestamp():
    now = datetime.datetime.utcnow()
    t = now.isoformat("T", "milliseconds")
    return t + "Z"


def signature(timestamp, method, request_path, body, secret_key):
    if str(body) == '{}' or str(body) == 'None':
        body = ''
    message = str(timestamp) + str.upper(method) + request_path + str(body)

    mac = hmac.new(bytes(secret_key, encoding='utf8'), bytes(message, encoding='utf-8'), digestmod='sha256')
    d = mac.digest()

    return base64.b64encode(d)


class OkxAPIException(Exception):

    def __init__(self, response):
        print(response.text + ', ' + str(response.status_code))
        self.code = 0
        try:
            json_res = response.json()
        except ValueError:
            self.message = 'Invalid JSON error message from Okx: {}'.format(response.text)
        else:
            if "code" in json_res.keys() and "msg" in json_res.keys():
                self.code = json_res['code']
                self.message = json_res['msg']
            else:
                self.code = 'None'
                self.message = 'System error'

        self.status_code = response.status_code
        self.response = response
        self.request = getattr(response, 'request', None)

    def __str__(self):  # pragma: no cover
        return 'API Request Error(code=%s): %s' % (self.code, self.message)


class OkxRequestException(Exception):

    def __init__(self, message):
        self.message = message

    def __str__(self):
        return 'OkxRequestException: %s' % self.message


class OkxParamsException(Exception):

    def __init__(self, message):
        self.message = message

    def __str__(self):
        return 'OkxParamsException: %s' % self.message


class Client(object):

    def __init__(self, api_key, api_secret_key, passphrase, sandbox, use_server_time=False, flag='1'):

        self.API_KEY = api_key
        self.API_SECRET_KEY = api_secret_key
        self.PASSPHRASE = passphrase
        self.use_server_time = use_server_time
        self.flag = flag
        self._sandbox = sandbox

        self.api_url = 'https://www.okx.com'

    def _request(self, method, request_path, params):

        if method == GET:
            request_path = request_path + parse_params_to_str(params)
        # url
        url = self.api_url + request_path

        timestamp = get_timestamp()

        # sign & header
        if self.use_server_time:
            timestamp = self._get_timestamp()

        body = json.dumps(params) if method == POST else ""

        sign_ = sign(pre_hash(timestamp, method, request_path, str(body)), self.API_SECRET_KEY)
        header = get_header(self.API_KEY, sign_, timestamp, self.PASSPHRASE, self.flag)
        if self._sandbox:
            header["x-simulated-trading"] = '1'

        # send request
        response = None

        if method == GET:
            response = requests.get(url, headers=header)
        elif method == POST:
            response = requests.post(url, data=body, headers=header)

        if not str(response.status_code).startswith('2'):
            raise OkxAPIException(response)

        return response.json()

    def _request_without_params(self, method, request_path):
        return self._request(method, request_path, {})

    def _request_with_params(self, method, request_path, params):
        return self._request(method, request_path, params)

    def _get_timestamp(self):
        url = self.api_url + SERVER_TIMESTAMP_URL
        response = requests.get(url)
        if response.status_code == 200:
            return response.json()['ts']
        else:
            return ""


class MarketAPI(Client):

    def __init__(self, api_key, api_secret_key, passphrase, sandbox, use_server_time=False, flag='1'):
        Client.__init__(self, api_key, api_secret_key, passphrase, sandbox, use_server_time, flag)

    # Get Tickers
    def get_tickers(self, instType, uly=''):
        if uly:
            params = {'instType': instType, 'uly': uly}
        else:
            params = {'instType': instType}
        return self._request_with_params(GET, TICKERS_INFO, params)

    # Get Ticker
    def get_ticker(self, instId):
        params = {'instId': instId}
        return self._request_with_params(GET, TICKER_INFO, params)

    # Get Index Tickers
    def get_index_ticker(self, quoteCcy='', instId=''):
        params = {'quoteCcy': quoteCcy, 'instId': instId}
        return self._request_with_params(GET, INDEX_TICKERS, params)

    # Get Order Book
    def get_orderbook(self, instId, sz=''):
        params = {'instId': instId, 'sz': sz}
        return self._request_with_params(GET, ORDER_BOOKS, params)

    # Get Candlesticks
    def get_candlesticks(self, instId, after='', before='', bar='', limit=''):
        params = {'instId': instId, 'after': after, 'before': before, 'bar': bar, 'limit': limit}
        return self._request_with_params(GET, MARKET_CANDLES, params)

    # GGet Candlesticks History（top currencies only）
    def get_history_candlesticks(self, instId, after='', before='', bar='', limit=''):
        params = {'instId': instId, 'after': after, 'before': before, 'bar': bar, 'limit': limit}
        return self._request_with_params(GET, HISTORY_CANDLES, params)

    # Get Index Candlesticks
    def get_index_candlesticks(self, instId, after='', before='', bar='', limit=''):
        params = {'instId': instId, 'after': after, 'before': before, 'bar': bar, 'limit': limit}
        return self._request_with_params(GET, INDEX_CANSLES, params)

    # Get Mark Price Candlesticks
    def get_markprice_candlesticks(self, instId, after='', before='', bar='', limit=''):
        params = {'instId': instId, 'after': after, 'before': before, 'bar': bar, 'limit': limit}
        return self._request_with_params(GET, MARKPRICE_CANDLES, params)

    # Get Index Candlesticks
    def get_trades(self, instId, limit=''):
        params = {'instId': instId, 'limit': limit}
        return self._request_with_params(GET, MARKET_TRADES, params)

    # Get Volume
    def get_volume(self):
        return self._request_without_params(GET, VOLUMNE)

    # Get Oracle
    def get_oracle(self):
        return self._request_without_params(GET, ORACLE)

    # Get Index Components
    def get_index_components(self, index):
        params = {'index': index}
        return self._request_with_params(GET, Components, params)

    # Get Tier
    def get_tier(self, instType='', tdMode='', uly='', instId='', ccy='', tier=''):
        params = {'instType': instType, 'tdMode': tdMode, 'uly': uly, 'instId': instId, 'ccy': ccy, 'tier': tier}
        return self._request_with_params(GET, TIER, params)

    # Get exchange rate
    def get_exchange_rate(self):
        params = {}
        return self._request_with_params(GET, BORROW_REPAY, params)


class TradeAPI(Client):

    def __init__(self, api_key, api_secret_key, passphrase, sandbox, use_server_time=False, flag='1'):
        Client.__init__(self, api_key, api_secret_key, passphrase, sandbox, use_server_time, flag)

    # Place Order
    def place_order(self, instId, tdMode, side, ordType, sz, ccy='', clOrdId='', tag='', posSide='', px='',
                    reduceOnly='', tgtCcy=''):
        params = {'instId': instId, 'tdMode': tdMode, 'side': side, 'ordType': ordType, 'sz': sz, 'ccy': ccy,
                  'clOrdId': clOrdId, 'tag': tag, 'posSide': posSide, 'px': px, 'reduceOnly': reduceOnly,
                  'tgtCcy': tgtCcy}
        return self._request_with_params(POST, PLACR_ORDER, params)

    # Place Multiple Orders
    def place_multiple_orders(self, orders_data):
        return self._request_with_params(POST, BATCH_ORDERS, orders_data)

    # Cancel Order
    def cancel_order(self, instId, ordId='', clOrdId=''):
        params = {'instId': instId, 'ordId': ordId, 'clOrdId': clOrdId}
        return self._request_with_params(POST, CANAEL_ORDER, params)

    # Cancel Multiple Orders
    def cancel_multiple_orders(self, orders_data):
        return self._request_with_params(POST, CANAEL_BATCH_ORDERS, orders_data)

    # Amend Order
    def amend_order(self, instId, cxlOnFail='', ordId='', clOrdId='', reqId='', newSz='', newPx=''):
        params = {'instId': instId, 'cxlOnFailc': cxlOnFail, 'ordId': ordId, 'clOrdId': clOrdId, 'reqId': reqId,
                  'newSz': newSz,
                  'newPx': newPx}
        return self._request_with_params(POST, AMEND_ORDER, params)

    # Amend Multiple Orders
    def amend_multiple_orders(self, orders_data):
        return self._request_with_params(POST, AMEND_BATCH_ORDER, orders_data)

    # Close Positions
    def close_positions(self, instId, mgnMode, posSide='', ccy='', autoCxl=''):
        params = {'instId': instId, 'mgnMode': mgnMode, 'posSide': posSide, 'ccy': ccy, 'autoCxl': autoCxl}
        return self._request_with_params(POST, CLOSE_POSITION, params)

    # Get Order Details
    def get_orders(self, instId, ordId='', clOrdId=''):
        params = {'instId': instId, 'ordId': ordId, 'clOrdId': clOrdId}
        return self._request_with_params(GET, ORDER_INFO, params)

    # Get Order List
    def get_order_list(self, instType='', uly='', instId='', ordType='', state='', after='', before='', limit=''):
        params = {'instType': instType, 'uly': uly, 'instId': instId, 'ordType': ordType, 'state': state,
                  'after': after, 'before': before, 'limit': limit}
        return self._request_with_params(GET, ORDERS_PENDING, params)

    # Get Order History (last 7 days）
    def get_orders_history(self, instType, uly='', instId='', ordType='', state='', after='', before='', limit=''):
        params = {'instType': instType, 'uly': uly, 'instId': instId, 'ordType': ordType, 'state': state,
                  'after': after, 'before': before, 'limit': limit}
        return self._request_with_params(GET, ORDERS_HISTORY, params)

    # Get Order History (last 3 months)
    def orders_history_archive(self, instType, uly='', instId='', ordType='', state='', after='', before='', limit=''):
        params = {'instType': instType, 'uly': uly, 'instId': instId, 'ordType': ordType, 'state': state,
                  'after': after, 'before': before, 'limit': limit}
        return self._request_with_params(GET, ORDERS_HISTORY_ARCHIVE, params)

    # Get Transaction Details
    def get_fills(self, instType='', uly='', instId='', ordId='', after='', before='', limit=''):
        params = {'instType': instType, 'uly': uly, 'instId': instId, 'ordId': ordId, 'after': after, 'before': before,
                  'limit': limit}
        return self._request_with_params(GET, ORDER_FILLS, params)

    # Place Algo Order
    def place_algo_order(self, instId='', tdMode='', side='', ordType='', sz='', ccy='',
                         posSide='', reduceOnly='', tpTriggerPx='',
                         tpOrdPx='', slTriggerPx='', slOrdPx='',
                         triggerPx='', orderPx='', tgtCcy='', pxVar='',
                         pxSpread='',
                         szLimit='', pxLimit='', timeInterval='', tpTriggerPxType='', slTriggerPxType='',
                         callbackRatio='', callbackSpread='', activePx='', tag='', triggerPxType=''):
        params = {'instId': instId, 'tdMode': tdMode, 'side': side, 'ordType': ordType, 'sz': sz, 'ccy': ccy,
                  'posSide': posSide, 'reduceOnly': reduceOnly, 'tpTriggerPx': tpTriggerPx, 'tpOrdPx': tpOrdPx,
                  'slTriggerPx': slTriggerPx, 'slOrdPx': slOrdPx, 'triggerPx': triggerPx, 'orderPx': orderPx,
                  'tgtCcy': tgtCcy, 'pxVar': pxVar, 'szLimit': szLimit, 'pxLimit': pxLimit,
                  'timeInterval': timeInterval,
                  'pxSpread': pxSpread, 'tpTriggerPxType': tpTriggerPxType, 'slTriggerPxType': slTriggerPxType,
                  'callbackRatio': callbackRatio, 'callbackSpread': callbackSpread, 'activePx': activePx,
                  'tag': tag, 'triggerPxType': triggerPxType, }
        return self._request_with_params(POST, PLACE_ALGO_ORDER, params)

    # Cancel Algo Order
    def cancel_algo_order(self, params):
        return self._request_with_params(POST, CANCEL_ALGOS, params)

    # Cancel Advance Algos
    def cancel_advance_algos(self, params):
        return self._request_with_params(POST, Cancel_Advance_Algos, params)

    # Get Algo Order List
    def order_algos_list(self, ordType, algoId='', instType='', instId='', after='', before='', limit=''):
        params = {'ordType': ordType, 'algoId': algoId, 'instType': instType, 'instId': instId, 'after': after,
                  'before': before, 'limit': limit}
        return self._request_with_params(GET, ORDERS_ALGO_OENDING, params)

    # Get Algo Order History
    def order_algos_history(self, ordType, state='', algoId='', instType='', instId='', after='', before='', limit=''):
        params = {'ordType': ordType, 'state': state, 'algoId': algoId, 'instType': instType, 'instId': instId,
                  'after': after, 'before': before, 'limit': limit}
        return self._request_with_params(GET, ORDERS_ALGO_HISTORY, params)

    # Get Transaction Details History
    def get_fills_history(self, instType, uly='', instId='', ordId='', after='', before='', limit=''):
        params = {'instType': instType, 'uly': uly, 'instId': instId, 'ordId': ordId, 'after': after, 'before': before,
                  'limit': limit}
        return self._request_with_params(GET, ORDERS_FILLS_HISTORY, params)


class ConvertAPI(Client):
    def __init__(self, api_key, api_secret_key, passphrase, sandbox, use_server_time=False, flag='1'):
        Client.__init__(self, api_key, api_secret_key, passphrase, sandbox, use_server_time, flag)

    def get_currencies(self):
        params = {}
        return self._request_with_params(GET, GET_CURRENCIES, params)

    def get_currency_pair(self, fromCcy='', toCcy=''):
        params = {"fromCcy": fromCcy, 'toCcy': toCcy}
        return self._request_with_params(GET, GET_CURRENCY_PAIR, params)

    def estimate_quote(self, baseCcy='', quoteCcy='', side='', rfqSz='', rfqSzCcy='', clQReqId=''):
        params = {'baseCcy': baseCcy, 'quoteCcy': quoteCcy, 'side': side, 'rfqSz': rfqSz, 'rfqSzCcy': rfqSzCcy,
                  'clQReqId': clQReqId}
        return self._request_with_params(POST, ESTIMATE_QUOTE, params)

    def convert_trade(self, quoteId='', baseCcy='', quoteCcy='', side='', sz='', szCcy='', clTReqId=''):
        params = {'quoteId': quoteId, 'baseCcy': baseCcy, 'quoteCcy': quoteCcy, 'side': side, 'sz': sz, 'szCcy': szCcy,
                  'clTReqId': clTReqId}
        return self._request_with_params(POST, CONVERT_TRADE, params)

    def get_convert_history(self, after='', before='', limit=''):
        params = {'after': after, 'before': before, 'limit': limit}
        return self._request_with_params(GET, CONVERT_HISTORY, params)


class AccountAPI(Client):

    def __init__(self, api_key, api_secret_key, passphrase, sandbox, use_server_time=False, flag='1'):
        Client.__init__(self, api_key, api_secret_key, passphrase, sandbox, use_server_time, flag)

    # Get Positions
    def get_position_risk(self, instType=''):
        params = {}
        if instType:
            params['instType'] = instType
        return self._request_with_params(GET, POSITION_RISK, params)

    # Get Balance
    def get_account(self, ccy=''):
        params = {}
        if ccy:
            params['ccy'] = ccy
        return self._request_with_params(GET, ACCOUNT_INFO, params)

    # Get Positions
    def get_positions(self, instType='', instId=''):
        params = {'instType': instType, 'instId': instId}
        return self._request_with_params(GET, POSITION_INFO, params)

    # Get Bills Details (recent 7 days)
    def get_bills_detail(self, instType='', ccy='', mgnMode='', ctType='', type='', subType='', after='', before='',
                         limit=''):
        params = {'instType': instType, 'ccy': ccy, 'mgnMode': mgnMode, 'ctType': ctType, 'type': type,
                  'subType': subType, 'after': after, 'before': before, 'limit': limit}
        return self._request_with_params(GET, BILLS_DETAIL, params)

    # Get Bills Details (recent 3 months)
    def get_bills_details(self, instType='', ccy='', mgnMode='', ctType='', type='', subType='', after='', before='',
                          limit=''):
        params = {'instType': instType, 'ccy': ccy, 'mgnMode': mgnMode, 'ctType': ctType, 'type': type,
                  'subType': subType, 'after': after, 'before': before, 'limit': limit}
        return self._request_with_params(GET, BILLS_ARCHIVE, params)

    # Get Account Configuration
    def get_account_config(self):
        return self._request_without_params(GET, ACCOUNT_CONFIG)

    # Get Account Configuration
    def get_position_mode(self, posMode):
        params = {'posMode': posMode}
        return self._request_with_params(POST, POSITION_MODE, params)

    # Get Account Configuration
    def set_leverage(self, lever, mgnMode, instId='', ccy='', posSide=''):
        params = {'lever': lever, 'mgnMode': mgnMode, 'instId': instId, 'ccy': ccy, 'posSide': posSide}
        return self._request_with_params(POST, SET_LEVERAGE, params)

    # Get Maximum Tradable Size For Instrument
    def get_maximum_trade_size(self, instId, tdMode, ccy='', px='', leverage=''):
        params = {'instId': instId, 'tdMode': tdMode, 'ccy': ccy, 'px': px, 'leverage': leverage}
        return self._request_with_params(GET, MAX_TRADE_SIZE, params)

    # Get Maximum Available Tradable Amount
    def get_max_avail_size(self, instId, tdMode, ccy='', reduceOnly=''):
        params = {'instId': instId, 'tdMode': tdMode, 'ccy': ccy, 'reduceOnly': reduceOnly}
        return self._request_with_params(GET, MAX_AVAIL_SIZE, params)

    # Increase / Decrease margin
    def Adjustment_margin(self, instId, posSide, type, amt, loanTrans=''):
        params = {'instId': instId, 'posSide': posSide, 'type': type, 'amt': amt, 'loanTrans': loanTrans}
        return self._request_with_params(POST, ADJUSTMENT_MARGIN, params)

    # Get Leverage
    def get_leverage(self, instId, mgnMode):
        params = {'instId': instId, 'mgnMode': mgnMode}
        return self._request_with_params(GET, GET_LEVERAGE, params)

    # Get the maximum loan of isolated MARGIN
    def get_max_load(self, instId, mgnMode, mgnCcy):
        params = {'instId': instId, 'mgnMode': mgnMode, 'mgnCcy': mgnCcy}
        return self._request_with_params(GET, MAX_LOAN, params)

    # Get Fee Rates
    def get_fee_rates(self, instType, instId='', uly='', category=''):
        params = {'instType': instType, 'instId': instId, 'uly': uly, 'category': category}
        return self._request_with_params(GET, FEE_RATES, params)

    # Get interest-accrued
    def get_interest_accrued(self, instId='', ccy='', mgnMode='', after='', before='', limit='', type=''):
        params = {'instId': instId, 'ccy': ccy, 'mgnMode': mgnMode, 'after': after, 'before': before, 'limit': limit,
                  'type': type}
        return self._request_with_params(GET, INTEREST_ACCRUED, params)

    # Get interest-accrued
    def get_interest_rate(self, ccy=''):
        params = {'ccy': ccy}
        return self._request_with_params(GET, INTEREST_RATE, params)

    # Set Greeks (PA/BS)
    def set_greeks(self, greeksType):
        params = {'greeksType': greeksType}
        return self._request_with_params(POST, SET_GREEKS, params)

    # Set Isolated Mode
    def set_isolated_mode(self, isoMode, type):
        params = {'isoMode': isoMode, 'type': type}
        return self._request_with_params(POST, ISOLATED_MODE, params)

    # Get Maximum Withdrawals
    def get_max_withdrawal(self, ccy=''):
        params = {'ccy': ccy}
        return self._request_with_params(GET, MAX_WITHDRAWAL, params)

    # Get account risk state
    def get_account_risk(self):
        return self._request_without_params(GET, ACCOUNT_RISK)

    # Get borrow repay
    def borrow_repay(self, ccy='', side='', amt=''):
        params = {'ccy': ccy, 'side': side, 'amt': amt}
        return self._request_with_params(POST, BORROW_REPAY, params)

    # Get borrow repay history
    def get_borrow_repay_history(self, ccy='', after='', before='', limit=''):
        params = {'ccy': ccy, 'after': after, 'before': before, 'limit': limit}
        return self._request_with_params(GET, BORROW_REPAY_HISTORY, params)

    # Get Obtain borrowing rate and limit
    def get_interest_limits(self, type='', ccy=''):
        params = {'type': type, 'ccy': ccy}
        return self._request_with_params(GET, INTEREST_LIMITS, params)

    # Get Simulated Margin
    def get_simulated_margin(self, instType='', inclRealPos='', instId='', pos=''):
        params = {'instType': instType, 'inclRealPos': inclRealPos, 'instId': instId, 'pos': pos, }
        return self._request_with_params(GET, SIMULATED_MARGIN, params)

    # Get  Greeks
    def get_greeks(self, ccy=''):
        params = {'ccy': ccy, }
        return self._request_with_params(GET, GREEKS, params)


class FundingAPI(Client):

    def __init__(self, api_key, api_secret_key, passphrase, sandbox, use_server_time=False, flag='1'):
        Client.__init__(self, api_key, api_secret_key, passphrase, sandbox, use_server_time, flag)

    # Get Deposit Address
    def get_deposit_address(self, ccy):
        params = {'ccy': ccy}
        return self._request_with_params(GET, DEPOSIT_ADDRESS, params)

    # Get Balance
    def get_balances(self, ccy=''):
        params = {'ccy': ccy}
        return self._request_with_params(GET, GET_BALANCES, params)

    # POST Account Configuration
    def funds_transfer(self, ccy, amt, froms, to, type='0', subAcct='', instId='', toInstId='', loanTrans=''):
        params = {'ccy': ccy, 'amt': amt, 'from': froms, 'to': to, 'type': type, 'subAcct': subAcct, 'instId': instId,
                  'toInstId': toInstId, 'loanTrans': loanTrans}
        return self._request_with_params(POST, FUNDS_TRANSFER, params)

    # Get Transfer State
    def transfer_state(self, transId, type=''):
        params = {'transId': transId, 'type': type}
        return self._request_with_params(POST, TRANSFER_STATE, params)

    # Withdrawal
    def coin_withdraw(self, ccy, amt, dest, toAddr, pwd, fee, chain=''):
        params = {'ccy': ccy, 'amt': amt, 'dest': dest, 'toAddr': toAddr, 'pwd': pwd, 'fee': fee, 'chain': chain}
        return self._request_with_params(POST, WITHDRAWAL_COIN, params)

    # Get Deposit History
    def get_deposit_history(self, ccy='', state='', after='', before='', limit='', txId=''):
        params = {'ccy': ccy, 'state': state, 'after': after, 'before': before, 'limit': limit, 'txId': txId}
        return self._request_with_params(GET, DEPOSIT_HISTORIY, params)

    # Get Withdrawal History
    def get_withdrawal_history(self, ccy='', state='', after='', before='', limit='', txId=''):
        params = {'ccy': ccy, 'state': state, 'after': after, 'before': before, 'limit': limit, 'txId': txId}
        return self._request_with_params(GET, WITHDRAWAL_HISTORIY, params)

    # Get Currencies
    def get_currency(self):
        return self._request_without_params(GET, CURRENCY_INFO)

    # PiggyBank Purchase/Redemption
    def purchase_redempt(self, ccy, amt, side, rate):
        params = {'ccy': ccy, 'amt': amt, 'side': side, 'rate': rate}
        return self._request_with_params(POST, PURCHASE_REDEMPT, params)

    # Get Withdrawal History
    def get_bills(self, ccy='', type='', after='', before='', limit=''):
        params = {'ccy': ccy, 'type': type, 'after': after, 'before': before, 'limit': limit}
        return self._request_with_params(GET, BILLS_INFO, params)

    # Get Piggy Balance
    def get_piggy_balance(self, ccy=''):
        params = {}
        if ccy:
            params = {'ccy': ccy}
        return self._request_with_params(GET, PIGGY_BALANCE, params)

    # Get Deposit Lightning
    def get_deposit_lightning(self, ccy, amt, to=""):
        params = {'ccy': ccy, 'amt': amt}
        if to:
            params = {'to': to}
        return self._request_with_params(GET, DEPOSIT_LIGHTNING, params)

    # Withdrawal Lightning
    def withdrawal_lightning(self, ccy, invoice, pwd):
        params = {'ccy': ccy, 'invoice': invoice, 'pwd': pwd}
        return self._request_with_params(POST, WITHDRAWAL_LIGHTNING, params)

    # GET Obtain account asset valuation
    def get_asset_valuation(self, ccy):
        params = {'ccy': ccy}
        return self._request_with_params(GET, ASSET_VALUATION, params)

    # POST SET LENDING RATE
    def set_lending_rate(self, ccy, rate):
        params = {'ccy': ccy, 'rate': rate}
        return self._request_with_params(POST, SET_LENDING_RATE, params)

    # GET LENDING HISTORY
    def get_lending_rate(self, ccy='', before='', after='', limit='', ):
        params = {'ccy': ccy, 'after': after, 'before': before, 'limit': limit, }
        return self._request_with_params(GET, LENDING_HISTORY, params)

    # GET LENDING RATE HISTORY
    def get_lending_rate_history(self, ccy='', ):
        params = {'ccy': ccy, }
        return self._request_with_params(GET, LENDING_RATE_HISTORY, params)

    # GET LENDING RATE SUMMARY
    def get_lending_rate_summary(self, ccy='', before='', after='', limit='', ):
        params = {'ccy': ccy, 'after': after, 'before': before, 'limit': limit, }
        return self._request_with_params(GET, LENDING_RATE_SUMMARY, params)


class PublicAPI(Client):

    def __init__(self, api_key, api_secret_key, passphrase, sandbox, use_server_time=False, flag='1'):
        Client.__init__(self, api_key, api_secret_key, passphrase, sandbox, use_server_time, flag)

    # Get Instruments
    def get_instruments(self, instType, uly='', instId=''):
        params = {'instType': instType, 'uly': uly, 'instId': instId}
        return self._request_with_params(GET, INSTRUMENT_INFO, params)

    # Get Delivery/Exercise History
    def get_deliver_history(self, instType, uly, after='', before='', limit=''):
        params = {'instType': instType, 'uly': uly, 'after': after, 'before': before, 'limit': limit}
        return self._request_with_params(GET, DELIVERY_EXERCISE, params)

    # Get Open Interest
    def get_open_interest(self, instType, uly='', instId=''):
        params = {'instType': instType, 'uly': uly, 'instId': instId}
        return self._request_with_params(GET, OPEN_INTEREST, params)

    # Get Funding Rate
    def get_funding_rate(self, instId):
        params = {'instId': instId}
        return self._request_with_params(GET, FUNDING_RATE, params)

    # Get Funding Rate History
    def funding_rate_history(self, instId, after='', before='', limit=''):
        params = {'instId': instId, 'after': after, 'before': before, 'limit': limit}
        return self._request_with_params(GET, FUNDING_RATE_HISTORY, params)

    # Get Limit Price
    def get_price_limit(self, instId):
        params = {'instId': instId}
        return self._request_with_params(GET, PRICE_LIMIT, params)

    # Get Option Market Data
    def get_opt_summary(self, uly, expTime=''):
        params = {'uly': uly, 'expTime': expTime}
        return self._request_with_params(GET, OPT_SUMMARY, params)

    # Get Estimated Delivery/Excercise Price
    def get_estimated_price(self, instId):
        params = {'instId': instId}
        return self._request_with_params(GET, ESTIMATED_PRICE, params)

    # Get Discount Rate And Interest-Free Quota
    def discount_interest_free_quota(self, ccy=''):
        params = {'ccy': ccy}
        return self._request_with_params(GET, DICCOUNT_INTETEST_INFO, params)

    # Get System Time
    def get_system_time(self):
        return self._request_without_params(GET, SYSTEM_TIME)

    # Get Liquidation Orders
    def get_liquidation_orders(self, instType, mgnMode='', instId='', ccy='', uly='', alias='', state='', before='',
                               after='', limit=''):
        params = {'instType': instType, 'mgnMode': mgnMode, 'instId': instId, 'ccy': ccy, 'uly': uly,
                  'alias': alias, 'state': state, 'before': before, 'after': after, 'limit': limit}
        return self._request_with_params(GET, LIQUIDATION_ORDERS, params)

    # Get Mark Price
    def get_mark_price(self, instType, uly='', instId=''):
        params = {'instType': instType, 'uly': uly, 'instId': instId}
        return self._request_with_params(GET, MARK_PRICE, params)

    # Get Tier
    def get_tier(self, instType, tdMode, uly='', instId='', ccy='', tier=''):
        params = {'instType': instType, 'tdMode': tdMode, 'uly': uly, 'instId': instId, 'ccy': ccy, 'tier': tier}
        return self._request_with_params(GET, TIER, params)

    # Get Interest Rate and Loan Quota
    def get_interest_loan(self):
        return self._request_without_params(GET, INTEREST_LOAN)

    # Get underlying

    def get_underlying(self, instType):
        params = {'instType': instType}
        return self._request_with_params(GET, UNDERLYING, params)

    # GET Obtain the privileged currency borrowing leverage rate and currency borrowing limit
    def get_vip_interest_rate_loan_quota(self):
        params = {}
        return self._request_with_params(GET, VIP_INTEREST_RATE_LOAN_QUOTA, params)



"""
    Coinbase Pro exchange definitions and setup
    Copyright (C) 2021  Emerson Dove

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from blankly.exchanges.exchange import Exchange
from blankly.exchanges.interfaces.paper_trade.paper_trade_interface import PaperTradeInterface


class PaperTrade(Exchange):
    def __init__(self, authenticated_exchange: Exchange, initial_account_values: dict = None):
        # Giving the preferences path as none allows us to create a default
        Exchange.__init__(self, "paper_trade", authenticated_exchange.portfolio_name, None)

        # This construction is unique because it doesn't make the interface super call for construction
        self.interface = PaperTradeInterface(authenticated_exchange.get_interface(),
                                             initial_account_values)
        self.calls = authenticated_exchange.calls

    def start_limit_order_watch(self):
        """
        This enables a thread that can watch & execute limit orders in the background of the exchange interface

        These throw warnings because they're typed to the ABC class
        """
        self.interface.start_paper_trade_watchdog()

    def stop_limit_order_watch(self):
        """
        This stops the thread that watches & executes limit orders

        These throw warnings because they're typed to the ABC class
        """
        self.interface.stop_paper_trade_watchdog()

    """
    Builds information about the symbol on this exchange by making particular API calls
    """

    def get_asset_state(self, symbol):
        """
        This determines the internal properties of the exchange block.
        Should be implemented per-class because it requires different types of interaction with each exchange.
        """
        # TODO Populate this with useful information
        return self.interface.get_account(symbol)

    def get_exchange_state(self):
        """
        Exchange state is the external properties for the exchange block
        """
        # TODO Populate this with useful information
        return self.interface.get_products()

    def get_direct_calls(self):
        return self.calls



"""
    Alpaca API creation definition.
    Copyright (C) 2021 Arun Annamalai

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import alpaca_trade_api

from blankly.exchanges.auth.auth_constructor import AuthConstructor

live_url = "https://api.alpaca.markets"
paper_url = "https://paper-api.alpaca.markets"


def create_alpaca_client(auth: AuthConstructor, sandbox_mode=True):
    if sandbox_mode:
        api_url = paper_url
    else:
        api_url = live_url

    return alpaca_trade_api.REST(auth.keys['API_KEY'], auth.keys['API_SECRET'], api_url, 'v2', raw_data=True)


# class API:
#     def __init__(self, auth: AuthConstructor, paper_trading=True):
#         if paper_trading:
#             self.__api_url = APCA_API_PAPER_URL
#         else:
#             self.__api_url = APCA_API_LIVE_URL
#
#         self.alp_client = alpaca_trade_api.REST(auth.keys['API_KEY'], auth.keys['API_SECRET'], self.__api_url, 'v2')
#
